from __future__ import print_function
import numpy as np
from Florence.Utils import insensitive
from warnings import warn

# BASE CLASS FOR ALL MATERIAL MODELS - SHOULD NOT BE USED DIRECTLY
class Material(object):
    """Base class for all material models"""

    def __init__(self, mtype, ndim, energy_type="internal_energy",
        lame_parameter_1=None, lame_parameter_2=None, poissons_ratio=None, youngs_modulus=None,
        shear_modulus=None, transverse_iso_youngs_modulus=None, transverse_iso_shear_modulus=None,
        bulk_modulus=None, density=None, permittivity=None, permeability=None,
        is_compressible=True, is_incompressible=False, is_nearly_incompressible=False,
        is_nonisotropic=True,is_anisotropic=False,is_transversely_isotropic=False, anisotropic_orientations=None,
        **kwargs):


        # SAFETY CHECKS
        if not isinstance(mtype, str):
            raise TypeError("Type of material model should be given as a string")
        if not isinstance(energy_type, str):
            raise TypeError("Material energy can either be 'internal_energy' or 'enthalpy'")

        self.energy_type = energy_type

        # MATERIAL CONSTANTS
        self.mu = lame_parameter_1
        self.lamb = lame_parameter_2
        self.nu = poissons_ratio
        self.E = youngs_modulus
        self.E_A = transverse_iso_youngs_modulus
        self.G_A = transverse_iso_shear_modulus
        self.K = bulk_modulus
        self.rho = density
        # if self.rho is None:
        #     self.rho = 0.0

        self.e = permittivity
        self.u = permeability


        # SET ALL THE OPTIONAL KEYWORD ARGUMENTS
        for i in kwargs.keys():
            if "__" not in i:
                setattr(self,i,kwargs[i])

        self.mtype = mtype
        self.ndim = ndim


        if 'elec' not in insensitive(self.mtype):
            if 'magnet' not in insensitive(self.mtype):
                self.nvar = self.ndim
        elif 'elec' in insensitive(self.mtype) and 'magnet' not in insensitive(self.mtype):
            self.nvar = self.ndim + 1
        elif 'elec' not in insensitive(self.mtype) and 'magnet' in insensitive(self.mtype):
            self.nvar = self.ndim + 1
        elif 'elec' in insensitive(self.mtype) and 'magnet' in insensitive(self.mtype):
            self.nvar = self.ndim + 2
        else:
            self.nvar = self.ndim

        self.H_Voigt = None

        if self.mu is None or self.lamb is None:
            if self.E is not None and self.nu is not None:
                self.GetLameParametersFromYoungsPoisson()
            # else:
            #     warn("You must set the material constants for problem")

        try:
            if self.mtype == 'LinearElastic' or \
                self.mtype == 'IncrementalLinearElastic':

                if self.ndim == 2:
                    self.H_Voigt = self.lamb*np.array([[1.,1.,0.],[1.,1.,0],[0.,0.,0.]]) +\
                     self.mu*np.array([[2.,0.,0.],[0.,2.,0],[0.,0.,1.]])
                else:
                    block_1 = np.zeros((6,6),dtype=np.float64); block_1[:3,:3] = np.ones((3,3))
                    block_2 = np.eye(6,6); block_2[0,0],block_2[1,1],block_2[2,2]=2.,2.,2.
                    self.H_Voigt = self.lamb*block_1 + self.mu*block_2
            else:
                if self.ndim == 2:
                    self.vIijIkl = np.array([[1.,1.,0.],[1.,1.,0],[0.,0.,0.]])
                    self.vIikIjl = np.array([[2.,0.,0.],[0.,2.,0],[0.,0.,1.]])
                else:
                    block_1 = np.zeros((6,6),dtype=np.float64); block_1[:3,:3] = np.ones((3,3))
                    block_2 = np.eye(6,6); block_2[0,0],block_2[1,1],block_2[2,2]=2.,2.,2.
                    self.vIijIkl = block_1
                    self.vIikIjl = block_2

                I = np.eye(self.ndim,self.ndim)
                self.Iijkl = np.einsum('ij,kl',I,I)
                self.Iikjl = np.einsum('ik,jl',I,I) + np.einsum('il,jk',I,I)

        except TypeError:
            # CATCH ONLY TypeError. OTHER MATERIAL CONSTANT RELATED ERRORS ARE SELF EXPLANATORY
            raise ValueError("Material constants for {} does not seem correct".format(self.mtype))

        if self.H_Voigt is not None:
            self.H_VoigtSize = self.H_Voigt.shape[0]

        self.is_compressible = is_compressible
        self.is_nearly_incompressible = is_nearly_incompressible
        self.is_incompressible = is_incompressible

        self.is_anisotropic = is_anisotropic
        self.is_transversely_isotropic = is_transversely_isotropic
        self.is_nonisotropic = is_nonisotropic
        self.anisotropic_orientations = anisotropic_orientations

        self.has_low_level_dispatcher = False




    def SetFibresOrientation(self,anisotropic_orientations):
        self.anisotropic_orientations = anisotropic_orientations


    def GetFibresOrientation(self, mesh, interior_orientation=None, plot=False):
        """Convenience function for computing anisotropic orientations of fibres
            in a transversely isotropic material.
            The orientation is computed based on the popular concept of reinforced composites
            where for the elements at the boundary, the fibres are perpendicular to the boundary
            edge/face

            input:
                mesh:                           [Mesh]
                interior_orientation:           [1D numpy.array or list] orientation of all interior
                                                fibres. Default is negative X-axis i.e [-1.,0.] for 2D
                                                and [-1.,0.,0.] for 3D
        """

        ndim = mesh.InferSpatialDimension()
        if self.ndim != ndim:
            raise ValueError('Mesh object and material model do not have the same spatial dimension')

        if self.ndim == 2:

            edge_elements = mesh.GetElementsWithBoundaryEdges()

            self.anisotropic_orientations = np.zeros((mesh.nelem,self.ndim),dtype=np.float64)
            for iedge in range(edge_elements.shape[0]):
                coords = mesh.points[mesh.edges[iedge,:],:]
                min_x = min(coords[0,0],coords[1,0])
                dist = (coords[0,0:]-coords[1,:])/np.linalg.norm(coords[0,0:]-coords[1,:])

                if min_x != coords[0,0]:
                    dist *= -1

                self.anisotropic_orientations[edge_elements[iedge],:] = dist

            if interior_orientation is None:
                interior_orientation = [-1.,0.]
            for i in range(mesh.nelem):
                if np.allclose(self.anisotropic_orientations[i,:],0.):
                    self.anisotropic_orientations[i,:] = interior_orientation


            if plot:

                Xs,Ys = [],[]
                for i in range(mesh.nelem):
                    x_avg = np.sum(mesh.points[mesh.elements[i,:],0])/mesh.points[mesh.elements[i,:],0].shape[0]
                    y_avg = np.sum(mesh.points[mesh.elements[i,:],1])/mesh.points[mesh.elements[i,:],1].shape[0]

                    Xs=np.append(Xs,x_avg)
                    Ys=np.append(Ys,y_avg)

                import matplotlib.pyplot as plt
                figure = plt.figure()
                q = plt.quiver(Xs, Ys, self.anisotropic_orientations[:,0],
                    self.anisotropic_orientations[:,1], color='Teal',
                           headlength=5,width=0.004)

                if mesh.element_type == "tri":
                    plt.triplot(mesh.points[:,0],mesh.points[:,1], mesh.elements[:,:3],color='k')
                else:
                    from Florence.MeshGeneration.NodeArrangement import NodeArrangementQuad
                    C = mesh.InferPolynomialDegree() - 1
                    reference_edges = NodeArrangementQuad(C)[0]
                    reference_edges = np.concatenate((reference_edges,reference_edges[:,1,None]),axis=1)
                    reference_edges = np.delete(reference_edges,1,1)

                    all_edge_elements = mesh.GetElementsEdgeNumberingQuad()
                    mesh.GetEdgesQuad()
                    x_edges = np.zeros((C+2,mesh.all_edges.shape[0]))
                    y_edges = np.zeros((C+2,mesh.all_edges.shape[0]))

                    BasesOneD = np.eye(2,2)
                    for iedge in range(mesh.all_edges.shape[0]):
                        ielem = all_edge_elements[iedge,0]
                        edge = mesh.elements[ielem,reference_edges[all_edge_elements[iedge,1],:]]
                        x_edges[:,iedge], y_edges[:,iedge] = mesh.points[edge,:].T

                    plt.plot(x_edges,y_edges,'-k')

                plt.axis('equal')
                plt.axis('off')
                plt.show()

        elif self.ndim == 3:

            face_elements = mesh.GetElementsWithBoundaryFaces()

            self.anisotropic_orientations = np.zeros((mesh.nelem,self.ndim),dtype=np.float64)

            for iface in range(face_elements.shape[0]):
                coords = mesh.points[mesh.faces[iface,:],:]
                min_x = min(coords[0,0], coords[1,0], coords[2,0])

                # ORIENTS THE FIBRE TO ONE OF THE EDGES OF THE FACE
                fibre = (coords[0,:]-coords[1,:])/np.linalg.norm(coords[0,:]-coords[1,:])

                if min_x != coords[0,0]:
                    fibre *= -1

                self.anisotropic_orientations[face_elements[iface],:] = fibre

            if interior_orientation is None:
                interior_orientation = [-1.,0.,0.]
            for i in range(mesh.nelem):
                if np.allclose(self.anisotropic_orientations[i,:],0.):
                    self.anisotropic_orientations[i,:] = interior_orientation

            if plot:

                # all_face_elements = mesh.GetElementsFaceNumbering()
                Xs = np.zeros(mesh.elements.shape[0])
                Ys = np.zeros(mesh.elements.shape[0])
                Zs = np.zeros(mesh.elements.shape[0])

                # divider = mesh.points[mesh.elements[0,:],0].shape[0]
                # for i in range(mesh.nelem):
                #     Xs[i] = np.sum(mesh.points[mesh.elements[i,:],0])/divider
                #     Ys[i] = np.sum(mesh.points[mesh.elements[i,:],1])/divider
                #     Zs[i] = np.sum(mesh.points[mesh.elements[i,:],2])/divider

                divider = mesh.points[mesh.faces[0,:],0].shape[0]
                for i in range(mesh.faces.shape[0]):
                    Xs[face_elements[i,0]] = np.sum(mesh.points[mesh.faces[i,:],0])/divider
                    Ys[face_elements[i,0]] = np.sum(mesh.points[mesh.faces[i,:],1])/divider
                    Zs[face_elements[i,0]] = np.sum(mesh.points[mesh.faces[i,:],2])/divider



                import os
                os.environ['ETS_TOOLKIT'] = 'qt4'
                from mayavi import mlab
                from Florence.PostProcessing import PostProcess

                if mesh.element_type == "tet":
                    tmesh = PostProcess.TessellateTets(mesh,np.zeros_like(mesh.points),
                        interpolation_degree=0)
                elif mesh.element_type == "hex":
                    tmesh = PostProcess.TessellateHexes(mesh,np.zeros_like(mesh.points),
                        interpolation_degree=0)

                figure = mlab.figure(bgcolor=(1,1,1),fgcolor=(1,1,1),size=(1000,800))

                mlab.quiver3d(Xs, Ys, Zs, self.anisotropic_orientations[:,0],
                    self.anisotropic_orientations[:,1], self.anisotropic_orientations[:,2],
                    color=(0.,128./255,128./255),line_width=2)

                src = mlab.pipeline.scalar_scatter(tmesh.x_edges.T.copy().flatten(),
                tmesh.y_edges.T.copy().flatten(), tmesh.z_edges.T.copy().flatten())
                src.mlab_source.dataset.lines = tmesh.connections
                lines = mlab.pipeline.stripper(src)
                h_edges = mlab.pipeline.surface(lines, color = (0,0,0), line_width=2)

                mlab.show()


    def Linearise(self,energy):
        """Linearises a material model, by dispatching invariants to self.LineariseInvariant"""
        pass


    def LineariseInvariant(self,invariant_to_linearise):
        """Give an invariant in the form of a dictionary in order to linearise it.

            input:
                invariant_to_linearise:         [dict] must contain the following keys:
                                                    invariant [str] the invariant to linearise
                                                    cofficient [str] a material coefficient
                                                    kinematics [str] for instance deformation gradient tensor F
                                                    constants [str] constants like kronecker delta d_ij

            Linearisation is always carried out with respect to the right Cauchy-Green tensor (C)


        >>> material = Material("MooneyRivlin",3)
        >>> material.Linearise({'invariant':'uC:I','coefficient':'u','kinematics':'C','constants':'I'})
        Cauchy stress:      2*u*I
        Spatial Hessian:        0

        """

        if not isinstance(invariant_to_linearise,dict):
            raise ValueError('invariant_to_linearise should be a dictionary')

        if 'invariant' not in invariant_to_linearise.keys():
            raise ValueError("invariant_to_linearise should have at least one key named 'invariant' with no spaces")

        strip_invariant = "".join(invariant_to_linearise['invariant'].split())

        if 'coefficient' not in invariant_to_linearise.keys():
            coefficient = ''
            invariant = strip_invariant
        else:
            coefficient = "".join(invariant_to_linearise['coefficient'].split())
            invariant = strip_invariant.split(coefficient)

        if len(invariant) > 1:
            if invariant[0] == '':
                if invariant[1][0] == '*':
                    invariant = invariant[1][1:]
            elif invariant[1] == '':
                if invariant[0][-1] == '*':
                    invariant = invariant[0][:-1]


        delta = u'\u03B4'
        delta = delta.encode('utf-8')

        if "C:I" in invariant or "F:F" in invariant or "trC" in invariant or "II_F" in invariant or "I_C" in invariant:
            cauchy = "2.0/J*"+coefficient+"*I"
            elasticity = "0"

        if "G:I" in invariant or "H:H" in invariant or "trG" in invariant or "II_H" in invariant or "I_G" in invariant:
            cauchy = "2.0/J*"+coefficient+"*(trace(b)*I-b)*b"
            elasticity = "4.0/J*"+coefficient+"*(b_ij*b_kl - b_ikb_jl)"

        if "lnJ" in invariant:
            cauchy = "2.0/J*"+coefficient+"*I"
            elasticity = "4.0/J*"+coefficient+"*"+delta+"_ik*"+delta+"_jl"

        if "(J-1)**2" in invariant:
            cauchy = "2.0*"+coefficient+"*(J-1)*I"
            elasticity = "2.0*"+coefficient+"*(2*J-1)"+delta+"_ij*"+delta+"_jk"+\
                "-4.0*"+coefficient+"*(J-1)"+delta+"_ik*"+delta+"_jl"

        if "NCN" in invariant or "FNFN" in invariant or "II_FN" in invariant:
            cauchy = "2.0/J*"+coefficient+"*(FN)_i(FN)_j"
            elasticity = "0"

        if "NGN" in invariant or "HNHN" in invariant or "II_HN" in invariant:
            cauchy = "2.0/J*"+coefficient+"*( ((HN)_k(HN)_k)*I -(HN)_i(HN)_j )"
            elasticity = "4.0/J*"+coefficient+"*( -"+delta+"_ij*(HN)_k(HN)_l +(HN)_i(HN)_j*"+\
            delta+"_kl" + "(HN)_m(HN)_m*"+delta+"_ij"+delta+"_kl" + "-(HN)_m(HN)_m*"+delta+"_ik"+delta+"_jl" +\
            delta+"_il"+"(HN)_j(HN)_k"+delta+"_jl"+"(HN)_i(HN)_k )"

        if "cauchy" not in locals() or "elasticity" not in locals():
            cauchy = "NIL"
            elasticity = "NIL"
            warn("I could not linearise the invariant %s" % invariant)

        print("Cauchy stress tensor:\t\t\t", cauchy)
        print("Spatial Hessian:\t\t\t\t", elasticity)

        return cauchy, elasticity


    def GetYoungsPoissonsFromLameParameters(self):

        assert self.mu != None
        assert self.lamb != None

        self.E = self.mu*(3.0*self.lamb + 2.*self.mu)/(self.lamb + self.mu)
        self.nu = self.lamb/2.0/(self.lamb + self.mu)

    def GetLameParametersFromYoungsPoisson(self):

        assert self.nu != None
        assert self.E != None

        self.lamb = self.E*self.nu/(1.+self.nu)/(1.-2.0*self.nu)
        self.mu = self.E/2./(1+self.nu)


    @property
    def Types(self):
        """Returns available material types"""
        import os
        pwd = os.path.dirname(os.path.realpath(__file__))
        list_of_materials = os.listdir(pwd)

        list_of_materials = [list_of_materials[i].split(".")[0] for i in range(len(list_of_materials))]
        list_of_materials = list(np.unique(list_of_materials))
        if "__init__" in list_of_materials:
            idx = list_of_materials.index("__init__")
            del list_of_materials[idx]

        return np.asarray(list_of_materials).reshape(-1,1)


    def GetType(self):
        """Get the type of material used"""
        if self.mtype is None:
            raise ValueError("You have not specified a material type. "
                "Call the 'Types' property for a list of available material models")
        return self.mtype


    def SetType(self,mtype):
        """Set the type of material to be used"""
        self.mtype = mtype


    def pprint(self):
        """Pretty print"""

        import pandas
        from copy import deepcopy
        Dict = deepcopy(self.__dict__)
        for key in Dict.keys():
            if Dict[key] is None:
                Dict[key] = np.NAN
            if isinstance(Dict[key],np.ndarray):
                del Dict[key]

        print(pandas.DataFrame(Dict,index=["Available parameters:"]))









import numpy as np
cimport numpy as np

ctypedef long long Integer
ctypedef unsigned long long UInteger
ctypedef double Real


cdef extern from "_LowLevelAssemblyPerfectLaplacian_.h" nogil:
    void _GlobalAssemblyPerfectLaplacian_(const Real *points,
                        const UInteger* elements,
                        const Real* Eulerp,
                        const Real* bases,
                        const Real* Jm,
                        const Real* AllGauss,
                        Integer ndim,
                        Integer ngauss,
                        Integer nelem,
                        Integer nodeperelem,
                        Integer nnode,
                        int *I_stiff,
                        int *J_stiff,
                        Real *V_stiff,
                        Real *e_tensor,
                        Integer is_hessian_symmetric
                        ) nogil


def _LowLevelAssemblyPerfectLaplacian_(fem_solver, function_space, formulation, mesh, material, Real[:,::1] Eulerx, Real[::1] Eulerp):

    # GET VARIABLES FOR DISPATCHING TO C
    cdef Integer ndim                       = formulation.ndim
    cdef Integer nvar                       = formulation.nvar
    cdef Integer ngauss                     = function_space.AllGauss.shape[0]
    cdef Integer nelem                      = mesh.nelem
    cdef Integer nodeperelem                = mesh.elements.shape[1]
    cdef Integer nnode                      = mesh.points.shape[0]

    cdef np.ndarray[UInteger,ndim=2, mode='c'] elements = mesh.elements
    cdef np.ndarray[Real,ndim=2, mode='c'] points       = mesh.points
    cdef np.ndarray[Real,ndim=2, mode='c'] bases        = function_space.Bases
    cdef np.ndarray[Real,ndim=3, mode='c'] Jm           = function_space.Jm
    cdef np.ndarray[Real,ndim=1, mode='c'] AllGauss     = function_space.AllGauss.flatten()

    cdef np.ndarray[int,ndim=1,mode='c'] I_stiff        = np.zeros(int((nvar*nodeperelem)**2*nelem),np.int32)
    cdef np.ndarray[int,ndim=1,mode='c'] J_stiff        = np.zeros(int((nvar*nodeperelem)**2*nelem),np.int32)
    cdef np.ndarray[Real,ndim=1,mode='c'] V_stiff       = np.zeros(int((nvar*nodeperelem)**2*nelem),np.float64)

    # NEGATIVE DEFINITE
    if material.e.shape[0] != ndim:
        raise ValueError("Permittivity tensor has to have a size of (ndim x ndim)")

    cdef np.ndarray[Real,ndim=2, mode='c'] e            = -material.e
    cdef Integer is_hessian_symmetric                   = np.allclose(material.e, material.e.T, atol=1e-8)

    _GlobalAssemblyPerfectLaplacian_(   &points[0,0],
                                        &elements[0,0],
                                        &Eulerp[0],
                                        &bases[0,0],
                                        &Jm[0,0,0],
                                        &AllGauss[0],
                                        ndim,
                                        ngauss,
                                        nelem,
                                        nodeperelem,
                                        nnode,
                                        &I_stiff[0],
                                        &J_stiff[0],
                                        &V_stiff[0],
                                        &e[0,0],
                                        is_hessian_symmetric
                                        )


    return I_stiff, J_stiff, V_stiff




from time import time
import numpy as np
from warnings import warn
import multiprocessing as MP
import imp

from .GetInteriorCoordinates import GetInteriorNodesCoordinates
import Florence.ParallelProcessing.parmap as parmap
from Florence.Tensor import itemfreq, makezero, unique2d, remove_duplicates_2D

#--------------------------------------------------------------------------------------------------------------------------#
# SUPPLEMENTARY FUNCTIONS
def ElementLoopTri(elem,elements,points,MeshType,eps,Neval):
    xycoord_higher = GetInteriorNodesCoordinates(points[elements[elem,:],:],MeshType,elem,eps,Neval)
    return xycoord_higher

def HighOrderMeshTri_SEMISTABLE(C, mesh, Decimals=10, equally_spaced=False, check_duplicates=True,
    Parallel=False, nCPU=1, ComputeAll=False):

    from Florence.FunctionSpace import Tri
    from Florence.QuadratureRules.FeketePointsTri import FeketePointsTri
    from Florence.MeshGeneration.NodeArrangement import NodeArrangementTri

    # SWITCH OFF MULTI-PROCESSING FOR SMALLER PROBLEMS WITHOUT GIVING A MESSAGE
    if (mesh.elements.shape[0] < 1000) and (C < 8):
        Parallel = False
        nCPU = 1

    if not equally_spaced:
        eps =  FeketePointsTri(C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        hpBases = Tri.hpNodal.hpBases
        Neval = np.zeros((3,eps.shape[0]),dtype=np.float64)
        for i in range(3,eps.shape[0]):
            Neval[:,i]  = hpBases(0,eps[i,0],eps[i,1],1)[0]
    else:
        from Florence.QuadratureRules.EquallySpacedPoints import EquallySpacedPointsTri
        eps =  EquallySpacedPointsTri(C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        hpBases = Tri.hpNodal.hpBases
        Neval = np.zeros((3,eps.shape[0]),dtype=np.float64)
        for i in range(3,eps.shape[0]):
            Neval[:,i]  = hpBases(0,eps[i,0],eps[i,1],Transform=1,EvalOpt=1,equally_spaced=True)[0]

    # THIS IS NECESSARY FOR REMOVING DUPLICATES
    makezero(Neval, tol=1e-12)

    nodeperelem = mesh.elements.shape[1]
    renodeperelem = int((C+2.)*(C+3.)/2.)
    left_over_nodes = renodeperelem - nodeperelem

    reelements = -1*np.ones((mesh.elements.shape[0],renodeperelem),dtype=np.int64)
    reelements[:,:3] = mesh.elements
    iesize = int( C*(C+5.)/2. )
    repoints = np.zeros((mesh.points.shape[0]+iesize*mesh.elements.shape[0],mesh.points.shape[1]),dtype=np.float64)
    # repoints[:mesh.points.shape[0],:]=mesh.points[:,:2]
    repoints[:mesh.points.shape[0],:]=mesh.points

    pshape0, pshape1 = mesh.points.shape[0], mesh.points.shape[1]
    repshape0, repshape1 = repoints.shape[0], repoints.shape[1]

    telements = time()

    xycoord_higher=[]; ParallelTuple1=[]
    if Parallel:
        # GET HIGHER ORDER COORDINATES - PARALLEL
        ParallelTuple1 = parmap.map(ElementLoopTri,np.arange(0,mesh.elements.shape[0]),mesh.elements,mesh.points,'tri',eps,
            Neval,pool=MP.Pool(processes=nCPU))

    # LOOP OVER ELEMENTS
    maxNode = np.max(reelements)
    for elem in range(0,mesh.elements.shape[0]):

        # GET HIGHER ORDER COORDINATES
        if Parallel:
            xycoord_higher = ParallelTuple1[elem]
        else:
            # xycoord =  mesh.points[mesh.elements[elem,:],:]
            xycoord_higher = GetInteriorNodesCoordinates(mesh.points[mesh.elements[elem,:],:],'tri',elem,eps,Neval)

        # EXPAND THE ELEMENT CONNECTIVITY
        newElements = np.arange(maxNode+1,maxNode+1+left_over_nodes)
        # reelements[elem,3:] = np.arange(maxNode+1,maxNode+1+left_over_nodes)
        reelements[elem,3:] = newElements
        maxNode = newElements[-1]

        repoints[mesh.points.shape[0]+elem*iesize:mesh.points.shape[0]+(elem+1)*iesize] = xycoord_higher[3:,:]

    telements = time()-telements

    #--------------------------------------------------------------------------------------
    # NOW REMOVE DUPLICATED POINTS
    tnodes = time()
    nnode_linear = mesh.points.shape[0]
    # KEEP ZEROFY ON, OTHERWISE YOU GET STRANGE BEHVAIOUR
    rounded_repoints = repoints[nnode_linear:,:].copy()
    makezero(rounded_repoints)
    rounded_repoints = np.round(rounded_repoints,decimals=Decimals)
    # flattened_repoints = np.ascontiguousarray(rounded_repoints).view(np.dtype((np.void,
        # rounded_repoints.dtype.itemsize * rounded_repoints.shape[1])))
    # _, idx_repoints, inv_repoints = np.unique(flattened_repoints,return_index=True,return_inverse=True)
    _, idx_repoints, inv_repoints = unique2d(rounded_repoints,order=False,
        consider_sort=False,return_index=True,return_inverse=True)
    del rounded_repoints

    idx_repoints = np.concatenate((np.arange(nnode_linear),idx_repoints+nnode_linear))
    repoints = repoints[idx_repoints,:]

    unique_reelements, inv_reelements = np.unique(reelements[:,3:],return_inverse=True)
    unique_reelements = unique_reelements[inv_repoints]
    reelements = unique_reelements[inv_reelements]
    reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem-3)
    reelements = np.concatenate((mesh.elements,reelements),axis=1)

    # SANITY CHECK FOR DUPLICATES
    #---------------------------------------------------------------------#
    # NOTE THAT THIS REMAPS THE ELEMENT CONNECTIVITY FOR THE WHOLE MESH
    # AND AS A RESULT THE FIRST FEW COLUMNS WOULD NO LONGER CORRESPOND TO
    # LINEAR CONNECTIVITY
    if check_duplicates:
        last_shape = repoints.shape[0]
        deci = int(Decimals)-2
        if Decimals < 6:
            deci = Decimals
        repoints, idx_repoints, inv_repoints = remove_duplicates_2D(repoints, decimals=deci)
        unique_reelements, inv_reelements = np.unique(reelements,return_inverse=True)
        unique_reelements = unique_reelements[inv_repoints]
        reelements = unique_reelements[inv_reelements]
        reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem)
        if last_shape != repoints.shape[0]:
            warn('Duplicated points generated in high order mesh. Lower the "Decimals". I have fixed it for now')
    #---------------------------------------------------------------------#

    tnodes = time() - tnodes
    #------------------------------------------------------------------------------------------


    # BUILD EDGES NOW
    #------------------------------------------------------------------------------------------
    tedges = time()

    edge_to_elements = mesh.GetElementsWithBoundaryEdgesTri()
    node_arranger = NodeArrangementTri(C)[0]
    reedges = np.zeros((mesh.edges.shape[0],C+2),dtype=np.int64)
    # for i in range(mesh.edges.shape[0]):
        # reedges[i,:] = reelements[edge_to_elements[i,0],node_arranger[edge_to_elements[i,1],:]]
    reedges = reelements[edge_to_elements[:,0][:,None],node_arranger[edge_to_elements[:,1],:]]

    tedges = time()-tedges
    #------------------------------------------------------------------------------------------

    class nmesh(object):
        # """Construct pMesh"""
        points = repoints
        elements = reelements
        edges = reedges
        faces = np.array([[],[]])
        nnode = repoints.shape[0]
        nelem = reelements.shape[0]
        info = 'tri'

    # print '\npMeshing timing:\n\t\tElement loop 1:\t '+str(telements)+' seconds\n\t\tNode loop:\t\t '+str(tnodes)+\
    #  ' seconds'+'\n\t\tElement loop 2:\t '+str(telements_2)+' seconds\n\t\tEdge loop:\t\t '+str(tedges)+' seconds\n'


    return nmesh



import os, sys
sys.path.insert(1,'/home/roman/Dropbox/florence')
import numpy as np
from math import sqrt
from numpy import einsum
from Florence.Tensor import Voigt, trace, makezero, makezero3d
from Florence.LegendreTransform import LegendreTransform
from Florence import *
from Florence.FiniteElements.ElementalMatrices.KinematicMeasures import KinematicMeasures



def check_dispatch():

    ndim=3
    ngauss = 5
    mu1,mu2,mu3,lamb = 1.,2.,3.,4.

    F = np.random.rand(ngauss,ndim,ndim)
    D = np.random.rand(ngauss,ndim)
    E = np.random.rand(ngauss,ndim)
    N = np.random.rand(100,ndim)
    StrainTensors = KinematicMeasures(F,"nonlinear")

    from _AnisotropicMooneyRivlin_1_ import KineticMeasures
    material = AnisotropicMooneyRivlin_1(ndim,mu1=mu1,mu2=mu2,mu3=mu3,lamb=lamb)
    material.anisotropic_orientations = N

    stress_py = np.zeros_like(F)
    D_py = np.zeros((ngauss,ndim,1))
    if ndim==3:
        hessian_py = np.zeros((ngauss,6,6))
    elif ndim==2:
        hessian_py = np.zeros((ngauss,3,3))

    for g in range(ngauss):
        stress_py[g,:,:] = material.CauchyStress(StrainTensors,elem=0,gcounter=g)
        hessian_py[g,:,:] = material.Hessian(StrainTensors,elem=0,gcounter=g)

    stress_cpp, hessian_cpp = KineticMeasures(material,F,N[0][:,None])
    makezero3d(stress_cpp)
    makezero3d(hessian_cpp)
    makezero3d(hessian_py)
    makezero3d(hessian_cpp)

    # print(stress_py - stress_cpp)
    # print(hessian_py - hessian_cpp)
    print(np.allclose(hessian_cpp,hessian_py))
    print(np.allclose(stress_cpp,stress_py))



def check_dispatch_electro():

    ndim=2
    ngauss = 5
    # mu1,mu2,mu3,lamb,eps_1,eps_2,eps_3,eps_e = 1.,1.,1.,1.,1.,1.,1.,1.
    mu1,mu2,mu3,lamb,eps_1,eps_2,eps_3,eps_e = 1.,2.,3.,4.,5.,6.,7.,9.
    material = IsotropicElectroMechanics_105(ndim,mu1=mu1,mu2=mu2,lamb=lamb,eps_1=eps_1,eps_2=eps_2)
    # material = IsotropicElectroMechanics_106(ndim,mu1=mu1,mu2=mu2,lamb=lamb,eps_1=eps_1,eps_2=eps_2)
    # material = AnisotropicMooneyRivlin_1(ndim,mu1=mu1,mu2=mu2,mu3=mu3,lamb=lamb)

    F = np.random.rand(ngauss,ndim,ndim)
    D = np.random.rand(ngauss,ndim)
    E = np.random.rand(ngauss,ndim); #E=np.zeros_like(E)
    N = np.random.rand(100,ndim)
    material.anisotropic_orientations = N
    StrainTensors = KinematicMeasures(F,"nonlinear")

    # from _AnisotropicMooneyRivlin_1_ import KineticMeasures
    # from _Piezoelectric_100_ import KineticMeasures
    # from _IsotropicElectroMechanics_106_ import KineticMeasures
    from _IsotropicElectroMechanics_105_ import KineticMeasures

    stress_py = np.zeros_like(F)
    D_py = np.zeros((ngauss,ndim,1))
    if ndim==3:
        hessian_py = np.zeros((ngauss,9,9))
    elif ndim==2:
        hessian_py = np.zeros((ngauss,5,5))

    for g in range(ngauss):
        # print E[g,:]
        D_py[g,:,:] = material.ElectricDisplacementx(StrainTensors,E[g,:],elem=0,gcounter=g)
        stress_py[g,:,:] = material.CauchyStress(StrainTensors,D_py[g,:,:],elem=0,gcounter=g)
        hessian_py[g,:,:] = material.Hessian(StrainTensors,D_py[g,:,:],elem=0,gcounter=g)

    # stress_cpp, hessian_cpp = KineticMeasures(material,F,N[0][:,None])
    D_cpp, stress_cpp, hessian_cpp = KineticMeasures(material,F,E)

    makezero3d(D_py)
    makezero3d(D_cpp)
    makezero3d(stress_cpp)
    makezero3d(hessian_cpp)
    makezero3d(hessian_py)
    makezero3d(hessian_cpp)

    # print(D_py - D_cpp)
    # print(stress_py - stress_cpp)
    # print(hessian_py - hessian_cpp)
    print(np.allclose(D_cpp,D_py))
    print(np.allclose(stress_cpp,stress_py))
    print(np.allclose(hessian_cpp,hessian_py))



# check_dispatch()
check_dispatch_electro()



import numpy as np
from time import time
from warnings import warn
from .GetInteriorCoordinates import GetInteriorNodesCoordinates
from Florence.Tensor import itemfreq, makezero, unique2d, remove_duplicates_2D


#---------------------------------------------------------------------------------------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------------------------#

def HighOrderMeshQuad(C, mesh, Decimals=10, equally_spaced=False, check_duplicates=True,
    Parallel=False, nCPU=1):

    from Florence.FunctionSpace import Quad, QuadES
    from Florence.QuadratureRules import GaussLobattoPointsQuad
    from Florence.QuadratureRules.EquallySpacedPoints import EquallySpacedPoints
    from Florence.MeshGeneration.NodeArrangement import NodeArrangementQuad

    if not equally_spaced:
        eps = GaussLobattoPointsQuad(C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        Neval = np.zeros((4,eps.shape[0]),dtype=np.float64)
        for i in range(0,eps.shape[0]):
            Neval[:,i] = Quad.LagrangeGaussLobatto(0,eps[i,0],eps[i,1],arrange=1)[:,0]
    else:
        eps = EquallySpacedPoints(3,C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        Neval = np.zeros((4,eps.shape[0]),dtype=np.float64)
        for i in range(0,eps.shape[0]):
            Neval[:,i] = QuadES.Lagrange(0,eps[i,0],eps[i,1],arrange=1)[:,0]
    makezero(Neval)

    nodeperelem = mesh.elements.shape[1]
    renodeperelem = int((C+2)**2)
    left_over_nodes = renodeperelem - nodeperelem

    reelements = -1*np.ones((mesh.elements.shape[0],renodeperelem),dtype=np.int64)
    reelements[:,:4] = mesh.elements
    iesize = int(4*C + C**2)
    repoints = np.zeros((mesh.points.shape[0]+iesize*mesh.elements.shape[0],mesh.points.shape[1]),dtype=np.float64)
    repoints[:mesh.points.shape[0],:]=mesh.points


    #--------------------------------------------------------------------------------------
    telements = time()

    xycoord_higher=[]; ParallelTuple1=[]
    if Parallel:
        # GET HIGHER ORDER COORDINATES - PARALLEL
        ParallelTuple1 = parmap.map(ElementLoopTri,np.arange(0,mesh.elements.shape[0]),mesh.elements,mesh.points,'quad',eps,
            Neval,pool=MP.Pool(processes=nCPU))

    # LOOP OVER ELEMENTS
    maxNode = np.max(reelements)
    for elem in range(0,mesh.elements.shape[0]):

        # GET HIGHER ORDER COORDINATES
        if Parallel:
            xycoord_higher = ParallelTuple1[elem]
        else:
            xycoord_higher = GetInteriorNodesCoordinates(mesh.points[mesh.elements[elem,:],:],'quad',elem,eps,Neval)

        # EXPAND THE ELEMENT CONNECTIVITY
        newElements = np.arange(maxNode+1,maxNode+1+left_over_nodes)
        # reelements[elem,3:] = np.arange(maxNode+1,maxNode+1+left_over_nodes)
        reelements[elem,4:] = newElements
        maxNode = newElements[-1]

        repoints[mesh.points.shape[0]+elem*iesize:mesh.points.shape[0]+(elem+1)*iesize] = xycoord_higher[4:,:]


    telements = time()-telements

    #--------------------------------------------------------------------------------------
    # NOW REMOVE DUPLICATED POINTS
    tnodes = time()
    nnode_linear = mesh.points.shape[0]
    # KEEP ZEROFY ON, OTHERWISE YOU GET STRANGE BEHVAIOUR
    rounded_repoints = repoints[nnode_linear:,:].copy()
    makezero(rounded_repoints)
    rounded_repoints = np.round(rounded_repoints,decimals=Decimals)
    _, idx_repoints, inv_repoints = unique2d(rounded_repoints,order=False,
        consider_sort=False,return_index=True,return_inverse=True)
    del rounded_repoints

    idx_repoints = np.concatenate((np.arange(nnode_linear),idx_repoints+nnode_linear))
    repoints = repoints[idx_repoints,:]

    unique_reelements, inv_reelements = np.unique(reelements[:,4:],return_inverse=True)
    unique_reelements = unique_reelements[inv_repoints]
    reelements = unique_reelements[inv_reelements]
    reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem-4)
    reelements = np.concatenate((mesh.elements,reelements),axis=1)


    # SANITY CHECK FOR DUPLICATES
    #---------------------------------------------------------------------#
    # NOTE THAT THIS REMAPS THE ELEMENT CONNECTIVITY FOR THE WHOLE MESH
    # AND AS A RESULT THE FIRST FEW COLUMNS WOULD NO LONGER CORRESPOND TO
    # LINEAR CONNECTIVITY
    if check_duplicates:
        last_shape = repoints.shape[0]
        deci = int(Decimals)-2
        if Decimals < 6:
            deci = Decimals
        repoints, idx_repoints, inv_repoints = remove_duplicates_2D(repoints, decimals=deci)
        unique_reelements, inv_reelements = np.unique(reelements,return_inverse=True)
        unique_reelements = unique_reelements[inv_repoints]
        reelements = unique_reelements[inv_reelements]
        reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem)
        if last_shape != repoints.shape[0]:
            warn('Duplicated points generated in high order mesh. Lower the "Decimals". I have fixed it for now')
    #---------------------------------------------------------------------#

    tnodes = time() - tnodes
    #------------------------------------------------------------------------------------------

    #------------------------------------------------------------------------------------------
    # BUILD EDGES NOW
    tedges = time()

    edge_to_elements = mesh.GetElementsWithBoundaryEdgesQuad()
    node_arranger = NodeArrangementQuad(C)[0]
    reedges = np.zeros((mesh.edges.shape[0],C+2),dtype=np.int64)
    reedges = reelements[edge_to_elements[:,0][:,None],node_arranger[edge_to_elements[:,1],:]]

    tedges = time()-tedges
    #------------------------------------------------------------------------------------------



    class nmesh(object):
        points = repoints
        elements = reelements
        edges = reedges
        faces = []
        nnode = repoints.shape[0]
        nelem = reelements.shape[0]
        info = 'quad'

    return nmesh


from __future__ import print_function
import gc, os, sys
import multiprocessing
from copy import deepcopy
from warnings import warn
from time import time
import numpy as np
from numpy.linalg import norm
import scipy as sp
from Florence.Utils import insensitive

from Florence.FiniteElements.Assembly import Assemble, AssembleExplicit
from Florence.PostProcessing import *
from Florence.Solver import LinearSolver
from Florence import Mesh


__all__ = ["LaplacianSolver"]


class LaplacianSolver(object):
    """Solver for linear and non-linear finite elements for Laplacian type operators.
    """

    def __init__(self, fem_solver=None):
        # super(LaplacianSolver, self).__init__(*args,**kwargs)
        if fem_solver is not None:
            self.__dict__.update(fem_solver.__dict__)
        # self.fem_solver = fem_solver
        self.__makeoutput__ = fem_solver.__makeoutput__


    def Solve(self, formulation=None, mesh=None,
        material=None, boundary_condition=None,
        function_spaces=None, solver=None, Eulerx=None, Eulerp=None):
        """Main solution routine for LaplacianSolver """

        # CHECK FOR ATTRIBUTE FOR LOWLEVEL ASSEMBLY
        if material.nature == "linear" and material.has_low_level_dispatcher and self.has_low_level_dispatcher:
            if hasattr(material,'e'):
                if material.e is None or isinstance(material.e, float):
                    if material.mtype == "IdealDielectric":
                        material.e = material.eps_1*np.eye(formulation.ndim, formulation.ndim)
                    else:
                        raise ValueError("For optimise=True, you need to provide the material permittivity tensor (ndimxndim)")
            else:
                raise ValueError("For optimise=True, you need to provide the material permittivity tensor (ndimxndim)")


        # INITIATE DATA FOR THE ANALYSIS
        NodalForces, Residual = np.zeros((mesh.points.shape[0]*formulation.nvar,1),dtype=np.float64), \
            np.zeros((mesh.points.shape[0]*formulation.nvar,1),dtype=np.float64)
        # SET NON-LINEAR PARAMETERS
        self.NRConvergence = { 'Increment_'+str(Increment) : [] for Increment in range(self.number_of_load_increments) }

        # ALLOCATE FOR SOLUTION FIELDS
        if self.save_frequency == 1:
            # TotalDisp = np.zeros((mesh.points.shape[0],self.number_of_load_increments),dtype=np.float32)
            TotalDisp = np.zeros((mesh.points.shape[0],self.number_of_load_increments),dtype=np.float64)
        else:
            TotalDisp = np.zeros((mesh.points.shape[0],
                int(self.number_of_load_increments/self.save_frequency)),dtype=np.float64)

        # PRE-ASSEMBLY
        print('Assembling the system and acquiring neccessary information for the analysis...')
        tAssembly=time()

        # APPLY DIRICHELT BOUNDARY CONDITIONS AND GET DIRICHLET RELATED FORCES
        boundary_condition.GetDirichletBoundaryConditions(formulation, mesh, material, solver, self)

        # ALLOCATE FOR GEOMETRY - GetDirichletBoundaryConditions CHANGES THE MESH
        # SO EULERX SHOULD BE ALLOCATED AFTERWARDS
        if Eulerx is None:
            Eulerx = np.copy(mesh.points)
        if Eulerp is None:
            Eulerp = np.zeros((mesh.points.shape[0]))

        # FIND PURE NEUMANN (EXTERNAL) NODAL FORCE VECTOR
        NeumannForces = boundary_condition.ComputeNeumannForces(mesh, material, function_spaces,
            compute_traction_forces=True, compute_body_forces=self.add_self_weight)

        # ASSEMBLE STIFFNESS MATRIX AND TRACTION FORCES FOR THE FIRST TIME
        if self.analysis_type == "static":
            K, TractionForces, _, _ = Assemble(self, function_spaces[0], formulation, mesh, material,
                Eulerx, Eulerp)
        else:
            pass

        if self.analysis_nature == 'nonlinear':
            print('Finished all pre-processing stage. Time elapsed was', time()-tAssembly, 'seconds')
        else:
            print('Finished the assembly stage. Time elapsed was', time()-tAssembly, 'seconds')


        if self.analysis_type != 'static':
            pass
        else:
            if self.iterative_technique == "newton_raphson" or self.iterative_technique == "modified_newton_raphson":
                TotalDisp = self.StaticSolver(function_spaces, formulation, solver,
                    K,NeumannForces,NodalForces,Residual,
                    mesh, TotalDisp, Eulerx, Eulerp, material, boundary_condition)


        return self.__makeoutput__(mesh, TotalDisp, formulation, function_spaces, material)



    def StaticSolver(self, function_spaces, formulation, solver, K,
            NeumannForces, NodalForces, Residual,
            mesh, TotalDisp, Eulerx, Eulerp, material, boundary_condition):

        LoadIncrement = self.number_of_load_increments
        LoadFactor = 1./LoadIncrement
        AppliedDirichletInc = np.zeros(boundary_condition.applied_dirichlet.shape[0],dtype=np.float64)

        for Increment in range(LoadIncrement):

            # CHECK ADAPTIVE LOAD FACTOR
            if self.load_factor is not None:
                LoadFactor = self.load_factor[Increment]

            # APPLY NEUMANN BOUNDARY CONDITIONS
            DeltaF = LoadFactor*NeumannForces
            NodalForces += DeltaF
            # OBRTAIN INCREMENTAL RESIDUAL - CONTRIBUTION FROM BOTH NEUMANN AND DIRICHLET
            Residual = -boundary_condition.ApplyDirichletGetReducedMatrices(K,Residual,
                boundary_condition.applied_dirichlet,LoadFactor=LoadFactor,only_residual=True)
            Residual -= DeltaF
            # GET THE INCREMENTAL DISPLACEMENT
            AppliedDirichletInc = LoadFactor*boundary_condition.applied_dirichlet

            t_increment = time()

            # LET NORM OF THE FIRST RESIDUAL BE THE NORM WITH RESPECT TO WHICH WE
            # HAVE TO CHECK THE CONVERGENCE OF NEWTON RAPHSON. TYPICALLY THIS IS
            # NORM OF NODAL FORCES
            if Increment==0:
                self.NormForces = np.linalg.norm(Residual)
                # AVOID DIVISION BY ZERO
                if np.isclose(self.NormForces,0.0):
                    self.NormForces = 1e-14

            self.norm_residual = np.linalg.norm(Residual)/self.NormForces

            if self.iterative_technique == "newton_raphson":
                Eulerx, Eulerp, K, Residual = self.NewtonRaphson(function_spaces, formulation, solver,
                    Increment, K, NodalForces, Residual, mesh, Eulerx, Eulerp,
                    material, boundary_condition, AppliedDirichletInc)
            elif self.iterative_technique == "modified_newton_raphson":
                Eulerx, Eulerp, K, Residual = self.ModifiedNewtonRaphson(function_spaces, formulation, solver,
                    Increment, K, NodalForces, Residual, mesh, Eulerx, Eulerp,
                    material, boundary_condition, AppliedDirichletInc)

            # UPDATE DISPLACEMENTS FOR THE CURRENT LOAD INCREMENT
            TotalDisp[:,Increment] = Eulerp

            # PRINT LOG IF ASKED FOR
            if self.print_incremental_log:
                dmesh = Mesh()
                dmesh.points = TotalDisp[:,:formulation.ndim,Increment]
                dmesh_bounds = dmesh.Bounds
                if formulation.fields == "electro_mechanics":
                    _bounds = np.zeros((2,formulation.nvar))
                    _bounds[:,:formulation.ndim] = dmesh_bounds
                    _bounds[:,-1] = [TotalDisp[:,-1,Increment].min(),TotalDisp[:,-1,Increment].max()]
                    print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),_bounds)
                else:
                    print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),dmesh_bounds)

            # SAVE INCREMENTAL SOLUTION IF ASKED FOR
            if self.save_incremental_solution:
                from scipy.io import savemat
                if self.incremental_solution_filename is not None:
                    savemat(self.incremental_solution_filename+"_"+str(Increment),{'solution':TotalDisp[:,:,Increment]},do_compression=True)
                else:
                    raise ValueError("No file name provided to save incremental solution")


            print('\nFinished Load increment', Increment, 'in', time()-t_increment, 'seconds')
            try:
                print('Norm of Residual is',
                    np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces), '\n')
            except RuntimeWarning:
                print("Invalid value encountered in norm of Newton-Raphson residual")

            # STORE THE INFORMATION IF NEWTON-RAPHSON FAILS
            if self.newton_raphson_failed_to_converge:
                solver.condA = np.NAN
                if Increment==0:
                    TotalDisp = TotalDisp.ravel()
                else:
                    TotalDisp = TotalDisp[:,:Increment]
                self.number_of_load_increments = Increment
                break

            # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
            if self.break_at_increment != -1 and self.break_at_increment is not None:
                if self.break_at_increment == Increment:
                    if self.break_at_increment < LoadIncrement - 1:
                        print("\nStopping at increment {} as specified\n\n".format(Increment))
                        TotalDisp = TotalDisp[:,:,:Increment]
                        self.number_of_load_increments = Increment
                    break

        # print(TotalDisp.shape)
        return TotalDisp


    def NewtonRaphson(self, function_spaces, formulation, solver,
        Increment, K, NodalForces, Residual, mesh, Eulerx, Eulerp, material,
        boundary_condition, AppliedDirichletInc):

        Tolerance = self.newton_raphson_tolerance
        LoadIncrement = self.number_of_load_increments
        Iter = 0


        # APPLY INCREMENTAL DIRICHLET PER LOAD STEP (THIS IS INCREMENTAL NOT ACCUMULATIVE)
        IncDirichlet = boundary_condition.UpdateFixDoFs(AppliedDirichletInc,
            K.shape[0],formulation.nvar)
        # UPDATE EULERIAN COORDINATE
        # Eulerx += IncDirichlet[:,:formulation.ndim]
        # Eulerp += IncDirichlet[:,-1]
        Eulerp += IncDirichlet[:,0]

        while self.norm_residual > Tolerance or Iter==0:
            # GET THE REDUCED SYSTEM OF EQUATIONS
            K_b, F_b = boundary_condition.GetReducedMatrices(K,Residual)[:2]

            # SOLVE THE SYSTEM
            sol = solver.Solve(K_b,-F_b)

            # GET ITERATIVE SOLUTION
            dU = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar)

            # UPDATE THE EULERIAN COMPONENTS
            # # UPDATE THE GEOMETRY
            # Eulerx += dU[:,:formulation.ndim]
            # # GET ITERATIVE ELECTRIC POTENTIAL
            # Eulerp += dU[:,-1]
            Eulerp += dU[:,0]

            # BREAK FROM HERE IF ANALYSIS IS LINEAR
            if material.nature == 'linear':
                break

            # RE-ASSEMBLE - COMPUTE STIFFNESS AND INTERNAL TRACTION FORCES
            K, TractionForces = Assemble(self, function_spaces[0], formulation, mesh, material,
                Eulerx,Eulerp)[:2]

            # FIND THE RESIDUAL
            Residual[boundary_condition.columns_in] = TractionForces[boundary_condition.columns_in] -\
                NodalForces[boundary_condition.columns_in]

            # SAVE THE NORM
            self.rel_norm_residual = la.norm(Residual[boundary_condition.columns_in])
            if Iter==0:
                self.NormForces = la.norm(Residual[boundary_condition.columns_in])
                # AVOID DIVISION BY ZERO
                if np.isclose(self.NormForces,0.0):
                    self.NormForces = 1e-14
            self.norm_residual = np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces)

            # SAVE THE NORM
            self.NRConvergence['Increment_'+str(Increment)] = np.append(self.NRConvergence['Increment_'+str(Increment)],\
                self.norm_residual)

            print("Iteration {} for increment {}.".format(Iter, Increment) +\
                " Residual (abs) {0:>16.7g}".format(self.rel_norm_residual),
                "\t Residual (rel) {0:>16.7g}".format(self.norm_residual))

            # BREAK BASED ON RELATIVE NORM
            if np.abs(self.rel_norm_residual) < Tolerance:
                break

            # BREAK BASED ON INCREMENTAL SOLUTION - KEEP IT AFTER UPDATE
            if norm(dU) <=  self.newton_raphson_solution_tolerance:
                print("Incremental solution within tolerance i.e. norm(dU): {}".format(norm(dU)))
                break

            # UPDATE ITERATION NUMBER
            Iter +=1

            if Iter==self.maximum_iteration_for_newton_raphson and formulation.fields == "electro_mechanics":
                # raise StopIteration("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
                warn("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
                self.newton_raphson_failed_to_converge = True
                break

            if Iter==self.maximum_iteration_for_newton_raphson:
                self.newton_raphson_failed_to_converge = True
                break
            if np.isnan(self.norm_residual) or self.norm_residual>1e06:
                self.newton_raphson_failed_to_converge = True
                break

            # USER DEFINED CRITERIA TO BREAK OUT OF NEWTON-RAPHSON
            if self.user_defined_break_func != None:
                if self.user_defined_break_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                    break

            # USER DEFINED CRITERIA TO STOP NEWTON-RAPHSON AND THE WHOLE ANALYSIS
            if self.user_defined_stop_func != None:
                if self.user_defined_stop_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                    self.newton_raphson_failed_to_converge = True
                    break


        return Eulerx, Eulerp, K, Residual





    def ModifiedNewtonRaphson(self, function_spaces, formulation, solver,
        Increment, K, NodalForces, Residual, mesh, Eulerx, Eulerp, material,
        boundary_condition, AppliedDirichletInc):

        from Florence.FiniteElements.Assembly import AssembleInternalTractionForces

        Tolerance = self.newton_raphson_tolerance
        LoadIncrement = self.number_of_load_increments
        Iter = 0


        # APPLY INCREMENTAL DIRICHLET PER LOAD STEP (THIS IS INCREMENTAL NOT ACCUMULATIVE)
        IncDirichlet = boundary_condition.UpdateFixDoFs(AppliedDirichletInc,
            K.shape[0],formulation.nvar)
        # UPDATE EULERIAN COORDINATE
        Eulerx += IncDirichlet[:,:formulation.ndim]
        Eulerp += IncDirichlet[:,-1]

        # ASSEMBLE STIFFNESS PER TIME STEP
        K, TractionForces = Assemble(self, function_spaces[0], formulation, mesh, material,
            Eulerx,Eulerp)[:2]

        while self.norm_residual > Tolerance or Iter==0:
            # GET THE REDUCED SYSTEM OF EQUATIONS
            K_b, F_b = boundary_condition.GetReducedMatrices(K,Residual)[:2]

            # SOLVE THE SYSTEM
            sol = solver.Solve(K_b,-F_b)

            # GET ITERATIVE SOLUTION
            dU = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar)

            # UPDATE THE EULERIAN COMPONENTS
            # UPDATE THE GEOMETRY
            Eulerx += dU[:,:formulation.ndim]
            # GET ITERATIVE ELECTRIC POTENTIAL
            Eulerp += dU[:,-1]

            # RE-ASSEMBLE - COMPUTE INTERNAL TRACTION FORCES
            TractionForces = AssembleInternalTractionForces(self, function_spaces[0], formulation, mesh, material,
                Eulerx,Eulerp)

            # FIND THE RESIDUAL
            Residual[boundary_condition.columns_in] = TractionForces[boundary_condition.columns_in] -\
                NodalForces[boundary_condition.columns_in]

            # SAVE THE NORM
            self.rel_norm_residual = la.norm(Residual[boundary_condition.columns_in])
            if Iter==0:
                self.NormForces = la.norm(Residual[boundary_condition.columns_in])
            self.norm_residual = np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces)

            # SAVE THE NORM
            self.NRConvergence['Increment_'+str(Increment)] = np.append(self.NRConvergence['Increment_'+str(Increment)],\
                self.norm_residual)

            print("Iteration {} for increment {}.".format(Iter, Increment) +\
                " Residual (abs) {0:>16.7g}".format(self.rel_norm_residual),
                "\t Residual (rel) {0:>16.7g}".format(self.norm_residual))

            # BREAK BASED ON RELATIVE NORM
            if np.abs(self.rel_norm_residual) < Tolerance:
                break

            # BREAK BASED ON INCREMENTAL SOLUTION - KEEP IT AFTER UPDATE
            if norm(dU) <=  self.newton_raphson_solution_tolerance:
                print("Incremental solution within tolerance i.e. norm(dU): {}".format(norm(dU)))
                break

            # UPDATE ITERATION NUMBER
            Iter +=1

            if Iter==self.maximum_iteration_for_newton_raphson and formulation.fields == "electro_mechanics":
                # raise StopIteration("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
                warn("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
                self.newton_raphson_failed_to_converge = True
                break

            if Iter==self.maximum_iteration_for_newton_raphson:
                self.newton_raphson_failed_to_converge = True
                break
            if np.isnan(self.norm_residual) or self.norm_residual>1e06:
                self.newton_raphson_failed_to_converge = True
                break

            # USER DEFINED CRITERIA TO BREAK OUT OF NEWTON-RAPHSON
            if self.user_defined_break_func != None:
                if self.user_defined_break_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                    break

            # USER DEFINED CRITERIA TO STOP NEWTON-RAPHSON AND THE WHOLE ANALYSIS
            if self.user_defined_stop_func != None:
                if self.user_defined_stop_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                    self.newton_raphson_failed_to_converge = True
                    break

        return Eulerx, Eulerp, K, Residual




import numpy as np
from .NumericIntegrator import GaussLobattoQuadrature
from Florence.MeshGeneration.NodeArrangement import NodeArrangementQuad, NodeArrangementHex

def GaussLobattoPoints1D(C):
    return GaussLobattoQuadrature(C+2)[0]


def GaussLobattoPointsQuad(C):

    xs = GaussLobattoQuadrature(C+2)[0]
    x,y = np.meshgrid(xs,xs)
    points = np.concatenate((x.flatten()[:,None],y.flatten()[:,None]),axis=1)
    node_aranger = NodeArrangementQuad(C)[2]
    return points[node_aranger,:]


def GaussLobattoPointsHex(C):

    xs = GaussLobattoQuadrature(C+2)[0]
    x,y,z = np.meshgrid(xs,xs,xs)
    points = np.concatenate((y.T.flatten()[:,None],x.T.flatten()[:,None],z.T.flatten()[:,None]),axis=1)

     # p = 1
     # [-1. -1. -1.]
     # [ 1. -1. -1.]
     # [ 1.  1. -1.]
     # [-1.  1. -1.]
     # [-1. -1.  1.]
     # [ 1. -1.  1.]
     # [ 1.  1.  1.]
     # [-1.  1.  1.]

    node_aranger = NodeArrangementHex(C)[2]
    return points[node_aranger,:]





import numpy as np
from .VariationalPrinciple import VariationalPrinciple
from Florence import QuadratureRule, FunctionSpace

from Florence.FiniteElements.LocalAssembly.KinematicMeasures import *
from Florence.FiniteElements.LocalAssembly._KinematicMeasures_ import _KinematicMeasures_
from .DisplacementApproachIndices import *
from ._ConstitutiveStiffnessDF_ import __ConstitutiveStiffnessIntegrandDF__
from ._TractionDF_ import __TractionIntegrandDF__
from Florence.Tensor import issymetric

__all__ = ["DisplacementFormulation"]

class DisplacementFormulation(VariationalPrinciple):

    def __init__(self, mesh, variables_order=(1,),
        quadrature_rules=None, quadrature_type=None, function_spaces=None, compute_post_quadrature=True,
        equally_spaced_bases=False):

        if mesh.element_type != "tet" and mesh.element_type != "tri" and \
            mesh.element_type != "quad" and mesh.element_type != "hex":
            raise NotImplementedError( type(self).__name__, "has not been implemented for", mesh.element_type, "elements")

        if isinstance(variables_order,int):
            self.variables_order = (self.variables_order,)
        self.variables_order = variables_order

        super(DisplacementFormulation, self).__init__(mesh,variables_order=self.variables_order,
            quadrature_type=quadrature_type,quadrature_rules=quadrature_rules,function_spaces=function_spaces,
            compute_post_quadrature=compute_post_quadrature)

        self.fields = "mechanics"
        self.nvar = self.ndim

        C = mesh.InferPolynomialDegree() - 1
        mesh.InferBoundaryElementType()

        if quadrature_rules == None and self.quadrature_rules == None:

            # OPTION FOR QUADRATURE TECHNIQUE FOR TRIS AND TETS
            optimal_quadrature = 3
            # is_flattened = True
            is_flattened = False

            if mesh.element_type == "tri" or mesh.element_type == "tet":
                norder = 2*C
                # TAKE CARE OF C=0 CASE
                if norder == 0:
                    norder = 1

                norder_post = 2*(C+1)
            else:
                norder = C+2
                norder_post = 2*(C+2)

            # GET QUADRATURE
            quadrature = QuadratureRule(optimal=optimal_quadrature, norder=norder, mesh_type=mesh.element_type, is_flattened=is_flattened)
            if self.compute_post_quadrature != None:
                # COMPUTE INTERPOLATION FUNCTIONS AT ALL INTEGRATION POINTS FOR POST-PROCESSING
                post_quadrature = QuadratureRule(optimal=optimal_quadrature, norder=norder_post, mesh_type=mesh.element_type)
            else:
                post_quadrature = None

            # BOUNDARY QUADRATURE
            bquadrature = QuadratureRule(optimal=optimal_quadrature, norder=C+2, mesh_type=mesh.boundary_element_type, is_flattened=is_flattened)

            self.quadrature_rules = (quadrature,post_quadrature,bquadrature)
        else:
            self.quadrature_rules = quadrature_rules

        if function_spaces == None and self.function_spaces == None:

            # CREATE FUNCTIONAL SPACES
            function_space = FunctionSpace(mesh, quadrature, p=C+1, equally_spaced=equally_spaced_bases, use_optimal_quadrature=is_flattened)
            if self.compute_post_quadrature != None:
                post_function_space = FunctionSpace(mesh, post_quadrature, p=C+1, equally_spaced=equally_spaced_bases)
            else:
                post_function_space = None

            # CREATE BOUNDARY FUNCTIONAL SPACES
            bfunction_space = FunctionSpace(mesh.CreateDummyLowerDimensionalMesh(),
                bquadrature, p=C+1, equally_spaced=equally_spaced_bases, use_optimal_quadrature=is_flattened)

            self.function_spaces = (function_space,post_function_space,bfunction_space)
        else:
            self.function_spaces = function_spaces

        # local_size = function_space.Bases.shape[0]*self.nvar
        local_size = self.function_spaces[0].Bases.shape[0]*self.nvar
        self.local_rows = np.repeat(np.arange(0,local_size),local_size,axis=0)
        self.local_columns = np.tile(np.arange(0,local_size),local_size)
        self.local_size = local_size

        # FOR MASS
        local_size_m = self.function_spaces[0].Bases.shape[0]*self.ndim
        self.local_rows_mass = np.repeat(np.arange(0,local_size_m),local_size_m,axis=0)
        self.local_columns_mass = np.tile(np.arange(0,local_size_m),local_size_m)
        self.local_size_m = local_size_m



    def GetElementalMatrices(self, elem, function_space, mesh, material, fem_solver, Eulerx, TotalPot):

        massel=[]; f = []
        # GET THE FIELDS AT THE ELEMENT LEVEL
        LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
        EulerElemCoords = Eulerx[mesh.elements[elem,:],:]

        # COMPUTE THE STIFFNESS MATRIX
        if material.has_low_level_dispatcher:
            stiffnessel, t = self.__GetLocalStiffness__(function_space,material,
                LagrangeElemCoords,EulerElemCoords,fem_solver,elem)
        else:
            stiffnessel, t = self.GetLocalStiffness(function_space,material,
                LagrangeElemCoords,EulerElemCoords,fem_solver,elem)

        I_mass_elem = []; J_mass_elem = []; V_mass_elem = []
        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed is False:
            # COMPUTE THE MASS MATRIX
            if material.has_low_level_dispatcher:
                massel = self.__GetLocalMass__(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)
            else:
                massel = self.GetLocalMass(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)


        I_stiff_elem, J_stiff_elem, V_stiff_elem = self.FindIndices(stiffnessel)
        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed is False:
            I_mass_elem, J_mass_elem, V_mass_elem = self.FindIndices(massel)

        return I_stiff_elem, J_stiff_elem, V_stiff_elem, t, f, I_mass_elem, J_mass_elem, V_mass_elem



    def GetElementalMatricesInVectorForm(self, elem, function_space, mesh, material, fem_solver, Eulerx, TotalPot):

        massel=[]; f = []
        # GET THE FIELDS AT THE ELEMENT LEVEL
        LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
        EulerElemCoords = Eulerx[mesh.elements[elem,:],:]

        # COMPUTE THE TRACTION VECTOR
        if material.has_low_level_dispatcher:
            t = self.__GetLocalTraction__(function_space,material,
            # t = self.GetLocalTraction(function_space,material,
                LagrangeElemCoords,EulerElemCoords,fem_solver,elem)
        else:
            t = self.GetLocalTraction(function_space,material,
                LagrangeElemCoords,EulerElemCoords,fem_solver,elem)

        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed is False:
            # COMPUTE THE MASS MATRIX
            if material.has_low_level_dispatcher:
                # massel = self.__GetLocalMass__(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)
                massel = self.__GetLocalMass_Efficient__(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)
            else:
                # massel = self.GetLocalMass(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)
                massel = self.GetLocalMass_Efficient(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)

            if fem_solver.analysis_subtype == "explicit" and fem_solver.mass_type == "lumped":
                massel = self.GetLumpedMass(massel)


        return t, f, massel



    def GetLocalStiffness(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem=0):
        """Get stiffness matrix of the system"""

        nvar = self.nvar
        ndim = self.ndim
        nodeperelem = function_space.Bases.shape[0]

        det = np.linalg.det
        inv = np.linalg.inv
        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        # ALLOCATE
        stiffness = np.zeros((nodeperelem*nvar,nodeperelem*nvar),dtype=np.float64)
        tractionforce = np.zeros((nodeperelem*nvar,1),dtype=np.float64)
        B = np.zeros((nodeperelem*nvar,material.H_VoigtSize),dtype=np.float64)

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # UPDATE/NO-UPDATE GEOMETRY
        if fem_solver.requires_geometry_update:
            # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
            ParentGradientx = np.einsum('ijk,jl->kil',Jm, EulerELemCoords)
            # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
            SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
            # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
            detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))
        else:
            # SPATIAL GRADIENT AND MATERIAL GRADIENT TENSORS ARE EQUAL
            SpatialGradient = np.einsum('ikj',MaterialGradient)
            # COMPUTE ONCE detJ
            detJ = np.einsum('i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)))


        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE THE HESSIAN AT THIS GAUSS POINT
            H_Voigt = material.Hessian(StrainTensors,None,elem,counter)

            # COMPUTE CAUCHY STRESS TENSOR
            CauchyStressTensor = []
            if fem_solver.requires_geometry_update:
                CauchyStressTensor = material.CauchyStress(StrainTensors,None,elem,counter)

            # COMPUTE THE TANGENT STIFFNESS MATRIX
            BDB_1, t = self.ConstitutiveStiffnessIntegrand(B, SpatialGradient[counter,:,:],
                CauchyStressTensor, H_Voigt, requires_geometry_update=fem_solver.requires_geometry_update)

            # COMPUTE GEOMETRIC STIFFNESS MATRIX
            if material.nature != "linear":
                BDB_1 += self.GeometricStiffnessIntegrand(SpatialGradient[counter,:,:],CauchyStressTensor)
            # INTEGRATE TRACTION FORCE
            if fem_solver.requires_geometry_update:
                tractionforce += t*detJ[counter]

            # INTEGRATE STIFFNESS
            stiffness += BDB_1*detJ[counter]


        return stiffness, tractionforce



    def __GetLocalStiffness__(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem=0):
        """Get stiffness matrix of the system"""

        # GET LOCAL KINEMATICS
        SpatialGradient, F, detJ = _KinematicMeasures_(function_space.Jm, function_space.AllGauss[:,0],
            LagrangeElemCoords, EulerELemCoords, fem_solver.requires_geometry_update)
        # COMPUTE WORK-CONJUGATES AND HESSIAN AT THIS GAUSS POINT
        CauchyStressTensor, H_Voigt = material.KineticMeasures(F,elem=elem)
        # COMPUTE LOCAL CONSTITUTIVE STIFFNESS AND TRACTION
        stiffness, tractionforce = __ConstitutiveStiffnessIntegrandDF__(SpatialGradient,
            CauchyStressTensor,H_Voigt,detJ,self.nvar,fem_solver.requires_geometry_update)
        # COMPUTE GEOMETRIC STIFFNESS
        if material.nature != "linear":
            stiffness += self.__GeometricStiffnessIntegrand__(SpatialGradient,CauchyStressTensor,detJ)

        return stiffness, tractionforce


    def ConstitutiveStiffnessIntegrand(self, B, SpatialGradient, CauchyStressTensor, H_Voigt,
        requires_geometry_update=True):
        """Applies to displacement based formulation"""

        SpatialGradient = SpatialGradient.T.copy()
        FillConstitutiveB(B,SpatialGradient,self.ndim,self.nvar)

        BDB = B.dot(H_Voigt.dot(B.T))

        t=np.zeros((B.shape[0],1))
        if requires_geometry_update:
            TotalTraction = GetTotalTraction(CauchyStressTensor)
            t = np.dot(B,TotalTraction)

        return BDB, t


    def GetLocalTraction(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem=0):
        """Get traction vector of the system"""

        nvar = self.nvar
        ndim = self.ndim
        nodeperelem = function_space.Bases.shape[0]

        det = np.linalg.det
        inv = np.linalg.inv
        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        # ALLOCATE
        tractionforce = np.zeros((nodeperelem*nvar,1),dtype=np.float64)
        B = np.zeros((nodeperelem*nvar,material.H_VoigtSize),dtype=np.float64)

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # UPDATE/NO-UPDATE GEOMETRY
        if fem_solver.requires_geometry_update:
            # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
            ParentGradientx = np.einsum('ijk,jl->kil',Jm, EulerELemCoords)
            # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
            SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
            # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
            detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))
        else:
            # SPATIAL GRADIENT AND MATERIAL GRADIENT TENSORS ARE EQUAL
            SpatialGradient = np.einsum('ikj',MaterialGradient)
            # COMPUTE ONCE detJ
            detJ = np.einsum('i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)))


        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE CAUCHY STRESS TENSOR
            CauchyStressTensor = []
            if fem_solver.requires_geometry_update:
                CauchyStressTensor = material.CauchyStress(StrainTensors,None,elem,counter)

            # COMPUTE THE TANGENT STIFFNESS MATRIX
            t = self.TractionIntegrand(B, SpatialGradient[counter,:,:],
                CauchyStressTensor, requires_geometry_update=fem_solver.requires_geometry_update)

            if fem_solver.requires_geometry_update:
                # INTEGRATE TRACTION FORCE
                tractionforce += t*detJ[counter]


        return tractionforce



    def __GetLocalTraction__(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem=0):
        """Get traction vector of the system"""

        # GET LOCAL KINEMATICS
        SpatialGradient, F, detJ = _KinematicMeasures_(function_space.Jm, function_space.AllGauss[:,0],
            LagrangeElemCoords, EulerELemCoords, fem_solver.requires_geometry_update)
        # COMPUTE WORK-CONJUGATES AND HESSIAN AT THIS GAUSS POINT
        CauchyStressTensor, _ = material.KineticMeasures(F,elem=elem)
        # COMPUTE LOCAL CONSTITUTIVE STIFFNESS AND TRACTION
        tractionforce = __TractionIntegrandDF__(SpatialGradient,
            CauchyStressTensor,detJ,material.H_VoigtSize,self.nvar,fem_solver.requires_geometry_update)

        return tractionforce


    def TractionIntegrand(self, B, SpatialGradient, CauchyStressTensor,
        requires_geometry_update=True):
        """Applies to displacement based formulation"""

        SpatialGradient = SpatialGradient.T.copy()
        FillConstitutiveB(B,SpatialGradient,self.ndim,self.nvar)

        t=np.zeros((B.shape[0],1))
        if requires_geometry_update:
            TotalTraction = GetTotalTraction(CauchyStressTensor)
            t = np.dot(B,TotalTraction)

        return t



    def GetLocalResidual(self):
        pass



    def GetEnergy(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem=0):
        """Get virtual energy of the system. For dynamic analysis this is handy for computing conservation of energy.
            The routine computes the global form of virtual internal energy i.e. integral of "W(C,G,C)"". This can be
            computed purely in a Lagrangian configuration.
        """

        nvar = self.nvar
        ndim = self.ndim
        nodeperelem = function_space.Bases.shape[0]

        det = np.linalg.det
        inv = np.linalg.inv
        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        internal_energy = 0.

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientx = np.einsum('ijk,jl->kil',Jm, EulerELemCoords)
        # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
        SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
        # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
        detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))

        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):
            # COMPUTE THE INTERNAL ENERGY AT THIS GAUSS POINT
            energy = material.InternalEnergy(StrainTensors,elem,counter)
            # INTEGRATE INTERNAL ENERGY
            internal_energy += energy*detJ[counter]

        return internal_energy




    def GetLinearMomentum(self, function_space, material, LagrangeElemCoords, EulerELemCoords, VelocityElem, fem_solver, elem=0):
        """Get linear momentum or virtual power of the system. For dynamic analysis this is handy for computing conservation of linear momentum.
            The routine computes the global form of virtual power i.e. integral of "P:Grad_0(V)"" where P is first Piola-Kirchhoff
            stress tensor and Grad_0(V) is the material gradient of velocity. Alternatively in update Lagrangian format this could be
            computed using "Sigma: Grad(V) J" where Sigma is the Cauchy stress tensor and Grad(V) is the spatial gradient of velocity.
            The latter approach is followed here
        """

        nvar = self.nvar
        ndim = self.ndim
        nodeperelem = function_space.Bases.shape[0]

        det = np.linalg.det
        inv = np.linalg.inv
        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        internal_power = 0.

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)
        # TIME DERIVATIVE OF F
        Fdot = np.einsum('ij,kli->kjl', VelocityElem, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientx = np.einsum('ijk,jl->kil',Jm, EulerELemCoords)
        # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
        SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
        # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
        detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))

        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):
            GradV = np.dot(Fdot[counter,:,:],np.linalg.inv(F[counter,:,:]))
            # COMPUTE CAUCHY
            CauchyStressTensor = material.CauchyStress(StrainTensors,None,elem,counter)
            # INTEGRATE INTERNAL VIRTUAL POWER
            internal_power += np.einsum('ij,ij',CauchyStressTensor,GradV)*detJ[counter]

        return internal_power


import numpy as np
cimport numpy as np

ctypedef long long Integer
ctypedef unsigned long long UInteger
ctypedef double Real


cdef extern from "_LowLevelAssemblyExplicit_DF_DPF_.h" nogil:
    void _GlobalAssemblyExplicit_DF_DPF_(const Real *points,
                        const UInteger* elements,
                        const Real* Eulerx,
                        const Real* Eulerp,
                        const Real* bases,
                        const Real* Jm,
                        const Real* AllGauss,
                        Integer ndim,
                        Integer nvar,
                        Integer ngauss,
                        Integer nelem,
                        Integer nodeperelem,
                        Integer nnode,
                        Integer H_VoigtSize,
                        Integer requires_geometry_update,
                        Real *T,
                        Integer is_dynamic,
                        Integer* local_rows_mass,
                        Integer* local_cols_mass,
                        int *I_mass,
                        int *J_mass,
                        Real *V_mass,
                        Real rho,
                        Real mu,
                        Real mu1,
                        Real mu2,
                        Real mu3,
                        Real mue,
                        Real lamb,
                        Real eps_1,
                        Real eps_2,
                        Real eps_3,
                        Real eps_e,
                        const Real *anisotropic_orientations,
                        int material_number,
                        int formulation_number
                        ) nogil


def _LowLevelAssemblyExplicit_DF_DPF_(function_space, formulation, mesh, material, Real[:,::1] Eulerx, Real[::1] Eulerp):

    # GET VARIABLES FOR DISPATCHING TO C
    cdef Integer ndim                       = formulation.ndim
    cdef Integer nvar                       = formulation.nvar
    cdef Integer ngauss                     = function_space.AllGauss.shape[0]
    cdef Integer nelem                      = mesh.nelem
    cdef Integer nodeperelem                = mesh.elements.shape[1]
    cdef Integer nnode                      = mesh.points.shape[0]
    cdef Integer H_VoigtSize                = material.H_VoigtSize

    cdef np.ndarray[UInteger,ndim=2, mode='c'] elements = mesh.elements
    cdef np.ndarray[Real,ndim=2, mode='c'] points       = mesh.points
    cdef np.ndarray[Real,ndim=2, mode='c'] bases        = function_space.Bases
    cdef np.ndarray[Real,ndim=3, mode='c'] Jm           = function_space.Jm
    cdef np.ndarray[Real,ndim=1, mode='c'] AllGauss     = function_space.AllGauss.flatten()

    cdef Integer requires_geometry_update               = True
    cdef Integer is_dynamic                             = False

    cdef np.ndarray[Integer,ndim=1,mode='c'] local_rows_mass        = formulation.local_rows_mass
    cdef np.ndarray[Integer,ndim=1,mode='c'] local_cols_mass        = formulation.local_columns_mass

    cdef np.ndarray[int,ndim=1,mode='c'] I_mass         = np.zeros(1,np.int32)
    cdef np.ndarray[int,ndim=1,mode='c'] J_mass         = np.zeros(1,np.int32)
    cdef np.ndarray[Real,ndim=1,mode='c'] V_mass        = np.zeros(1,np.float64)

    if is_dynamic:
        I_mass          = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
        J_mass          = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
        V_mass          = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)


    cdef np.ndarray[Real,ndim=1,mode='c'] T = np.zeros(mesh.points.shape[0]*nvar,np.float64)

    cdef np.ndarray[Real,ndim=2,mode='c'] anisotropic_orientations = np.zeros((1,1),np.float64)
    if material.is_transversely_isotropic:
        anisotropic_orientations = material.anisotropic_orientations

    cdef Real mu=0.,mu1=0.,mu2=0.,mu3=0.,mue=0.,lamb=0.,eps_1=0.,eps_2=0., eps_3=0., eps_e=0.

    cdef int material_number
    if material.mtype == "ExplicitMooneyRivlin":
        mu1, mu2, lamb = material.mu1, material.mu2, material.lamb
        material_number = 0
    elif material.mtype == "NeoHookean":
        mu, lamb = material.mu, material.lamb
        material_number = 1
    elif material.mtype == "MooneyRivlin":
        mu1, mu2, lamb = material.mu1, material.mu2, material.lamb
        material_number = 2
    elif material.mtype == "NearlyIncompressibleMooneyRivlin":
        mu1, mu2, mu3 = material.alpha, material.beta, material.kappa
        material_number = 3
    elif material.mtype == "IsotropicElectroMechanics_101":
        mu, lamb, eps_1 = material.mu, material.lamb, material.eps_1
        material_number = 4
    elif material.mtype == "IsotropicElectroMechanics_105":
        mu1, mu2, lamb, eps_1, eps_2 = material.mu1, material.mu2, material.lamb, material.eps_1, material.eps_2
        material_number = 5
    elif material.mtype == "IsotropicElectroMechanics_106":
        mu1, mu2, lamb, eps_1, eps_2 = material.mu1, material.mu2, material.lamb, material.eps_1, material.eps_2
        material_number = 6
    elif material.mtype == "IsotropicElectroMechanics_107":
        mu1, mu2, mue, lamb, eps_1, eps_2, eps_e = material.mu1, material.mu2, material.mue, material.lamb, \
            material.eps_1, material.eps_2, material.eps_e
        material_number = 7
    elif material.mtype == "IsotropicElectroMechanics_108":
        mu1, mu2, lamb, eps_2 = material.mu1, material.mu2, material.lamb, material.eps_2
        material_number = 8
    elif material.mtype == "ExplicitIsotropicElectroMechanics_108":
        mu1, mu2, lamb, eps_2 = material.mu1, material.mu2, material.lamb, material.eps_2
        material_number = 9
    elif material.mtype == "LinearElastic" or material.mtype == "IncrementalLinearElastic":
        mu, lamb = material.mu, material.lamb
        material_number = 10
    else:
        raise NotImplementedError("Low level assembly for material {} not available for explicit analysis."
            " Consider 'optimise=False' for now".format(material.mtype))

    cdef Real rho = material.rho

    cdef int formulation_number
    if formulation.fields == "mechanics":
        formulation_number = 0
    elif formulation.fields == "electro_mechanics":
        formulation_number = 1

    _GlobalAssemblyExplicit_DF_DPF_(    &points[0,0],
                                        &elements[0,0],
                                        &Eulerx[0,0],
                                        &Eulerp[0],
                                        &bases[0,0],
                                        &Jm[0,0,0],
                                        &AllGauss[0],
                                        ndim,
                                        nvar,
                                        ngauss,
                                        nelem,
                                        nodeperelem,
                                        nnode,
                                        H_VoigtSize,
                                        requires_geometry_update,
                                        &T[0],
                                        is_dynamic,
                                        &local_rows_mass[0],
                                        &local_cols_mass[0],
                                        &I_mass[0],
                                        &J_mass[0],
                                        &V_mass[0],
                                        rho,
                                        mu,
                                        mu1,
                                        mu2,
                                        mu3,
                                        mue,
                                        lamb,
                                        eps_1,
                                        eps_2,
                                        eps_3,
                                        eps_e,
                                        &anisotropic_orientations[0,0],
                                        material_number,
                                        formulation_number
                                        )


    return I_mass, J_mass, V_mass, T






from libc.stdlib cimport malloc, free 
from cython cimport double, sizeof, boundscheck, wraparound

# cdef extern from "jacobi.c" nogil:
#     void jacobi(const unsigned short n, const double xi,
#             const double a, const double b, double *P)
#     void diffjacobi(const unsigned short n, const double xi,
#             const double a, const double b, const unsigned short opt, double *dP)


cdef inline void jacobi(const unsigned short n, const double xi, const double a, const double b, double *P) nogil:
    cdef double a1n,a2n,a3n,a4n;
    cdef unsigned short p;

    P[0]=1.0
    if n>0:
        P[1] = 0.5*((a-b)+(a+b+2)*xi)
    if n>1:
        for p in range(1,n):
            a1n = 2*(p+1)*(p+a+b+1)*(2*p+a+b)
            a2n = (2*p+a+b+1)*(a*a-b*b)
            a3n = (2*p+a+b)*(2*p+a+b+1)*(2*p+a+b+2)
            a4n = 2*(p+a)*(p+b)*(2*p+a+b+2)

            P[p+1] = ((a2n+a3n*xi)*P[p]-a4n*P[p-1])/a1n


cdef inline void diffjacobi(const unsigned short n, const double xi, const double a, const double b, const unsigned short opt, double *dP) nogil:
    cdef unsigned short p
    cdef double *P = <double*>malloc( (n+1)*sizeof(double))
    if opt==1:
        jacobi(n,xi,a+1,b+1,P)
    else:
        jacobi(n,xi,a,b,P)

    for p in range(1,n+1):
        dP[p] = 0.5*(a+b+p+1)*P[p-1]

    free(P)


@boundscheck(False)
@wraparound(False)
def JacobiPolynomials(const unsigned short n, double xi, double a=0., double b=0.):
    cdef:
        int i
        double *P = <double*>malloc( (n+1)*sizeof(double))
    jacobi(n,xi,a,b,P);
    P_py=[0]*(n+1)
    for i in range(n+1):
        P_py[i] = P[i]

    free(P)
    return P_py

@boundscheck(False)
@wraparound(False)
def DiffJacobiPolynomials(const int n,double xi,double a=0.,double b=0.,int opt=0):
    cdef:
        int i
        double *dP = <double*>malloc( (n+1)*sizeof(double))
    
    diffjacobi(n,xi,a,b,opt,dP);
    dP_py=[0]*(n+1)
    for i in range(n+1):
        dP_py[i] = dP[i]
    
    free(dP);
    return dP_py





#======================================================================#
#======================================================================#
#======================================================================#

from libc cimport math 
import numpy as np
cimport numpy as np



# 1D - LINE
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#

@boundscheck(False)
def NormalisedJacobi1D(int C, double x):

    cdef:
        double[::1] p = np.zeros(C+2)
        int i

    for i in range(0,C+2):
        p[i] = JacobiPolynomials(i,x,0,0)[-1]*math.sqrt((2.*i+1.)/2.)

    return p  


# 2D - TRI
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#

@boundscheck(False)
def NormalisedJacobi2D(int C, x):
    """Computes the orthogonal base of 2D polynomials of degree less than
        or equal to C+1 at the point x=(r,s) in [-1,1]**2 (i.e. on the reference quad)"""


    cdef:
        int N = (C+2)*(C+3)/2 
        double[::1] p = np.zeros(N)
        double r = x[0]
        double s = x[1]
        int nDeg, i, j
        double p_i, p_j, q_i, q_j

    # ORDERING: FIRST INCREASING THE DEGREE AND THEN LEXOGRAPHIC ORDER
    cdef int ncount = 0 
    # LOOP ON DEGREE
    for nDeg in range(0,C+2):
      # LOOP INCREASING I
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1.
                q_i = 1.  
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1]
                q_i = q_i*(1.-s)/2.
            # VALUE FOR J
            j = nDeg-i
            if j==0:
               p_j = 1.
            else:
               p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1]
            
            # factor = np.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            factor = math.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            p[ncount] = ( p_i*q_i*p_j )*factor

            ncount += 1
    return p 



def NormalisedJacobiTri(int C, x):   
    """Computes the orthogonal base of 2D polynomials of degree less than
        or equal to n at the point x=(xi,eta) in the reference triangle"""

    cdef:
        double xi = x[0]
        double eta = x[1]
        double r, s 

    if eta==1.: 
        r = -1.
        s = 1.
    else:
        r = 2.*(1+xi)/(1.-eta)-1.
        s = eta

    # return NormalisedJacobi2D(C,np.array([r,s]))
    return NormalisedJacobi2D(C,[r,s]) 


@boundscheck(False)
def GradNormalisedJacobiTri(int C,x, EvalOpt=0):
    """Computes the orthogonal base of 2D polynomials of degree less than
        or equal to n at the point x=(r,s) in [-1,1]**2"""
    
    cdef:
        int N = (C+2)*(C+3)/2 
        double[::1] p = np.zeros(N);
        double[::1] dp_dxi  = np.zeros(N)
        double[::1] dp_deta = np.zeros(N)
        double r = x[0]
        double s = x[1]
        double xi, eta, dr_dxi, dr_deta
        double dp_dr, dp_ds
        int nDeg, i, j
        double factor 

    # THIS MAY RUIN THE CONVERGENCE, BUT FOR POST PROCESSING ITS FINE
    if EvalOpt==1:
        if s==1:
            s=0.99999999999999

    xi = (1.+r)*(1.-s)/2.-1
    eta = s

    dr_dxi  = 2./(1.-eta)
    dr_deta = 2.*(1.+xi)/(1.-eta)**2
    # DERIVATIVE OF s IS NOT NEEDED SINCE s=eta

    # ORDERING: FIRST INCREASING THE DEGREE AND THEN LEXOGRAPHIC ORDER
    cdef int ncount = 0
    # Loop on degree
    for nDeg in range(0,C+2):
      # LOOP INCREASING I
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1.  
                q_i = 1.  
                dp_i = 0.
                dq_i = 0.
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1] 
                dp_i = JacobiPolynomials(i-1,r,1.,1.)[-1]*(i+1.)/2.    
                
                q_i = q_i*(1.-s)/2.
                dq_i = 1.*q_i*(-i)/(1-s)
            
            # VALUE FOR J
            j = nDeg-i
            if j==0:
                p_j = 1.
                dp_j = 0.
            else:
                p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1]
                dp_j = JacobiPolynomials(j-1,s,2.*i+2.,1.)[-1]*(j+2.*i+2.)/2.  
            
            # factor = np.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            factor = math.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            # NORMALISED POLYNOMIAL
            p[ncount] = ( p_i*q_i*p_j )*factor
            # DERIVATIVE WITH RESPECT TO (r,s)
            dp_dr = ( (dp_i)*q_i*p_j )*factor
            dp_ds = ( p_i*(dq_i*p_j+q_i*dp_j) )*factor
            # DERIVATIVE WITH RESPECT TO (xi,eta)
            dp_dxi[ncount]  = dp_dr*dr_dxi
            dp_deta[ncount] = dp_dr*dr_deta + dp_ds

            ncount += 1

    return p,dp_dxi,dp_deta



# 3D - TET
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#

@boundscheck(False)
def NormalisedJacobi3D(int C, x):

    """Computes the orthogonal base of 3D polynomials of degree less than
        or equal to n at the point x=(r,s,t) in [-1,1]**3
    """

    cdef int N = (C+2)*(C+3)*(C+4)/6
    cdef double[::1] p = np.zeros(N)

    cdef:
        double r = x[0]
        double s = x[1]
        double t = x[2]
        int nDeg, i, j, k
        double p_i, p_j, p_k, q_i, q_j, q_k
        double factor 

    # ORDERING: FIRST INCREASING THE DEGREE AND THEN LEXOGRAPHIC ORDER
    cdef int ncount = 0
    # LOOP ON DEGREE
    for nDeg in range(0,C+2):
        # LOOP INCREASING I
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1;  q_i = 1 
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1]
                q_i = q_i*(1.-s)/2.
            # LOOP INCREASING J
            for j in range(0,nDeg-i+1):
                if j==0:
                    p_j = 1
                    q_j = ((1.-t)/2.)**i
                else:
                    p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1]
                    q_j = q_j*(1.-t)/2.
                # VALUE FOR K
                k = nDeg-(i+j)
                if k==0:
                   p_k = 1.
                else:
                    p_k = JacobiPolynomials(k,t,2.*(i+j)+2.,0.)[-1]
                factor = math.sqrt( (2.*i+1.)*(i+j+1.)*(2.*(i+j+k)+3.)/4. )
                p[ncount] = ( p_i*q_i*p_j*q_j*p_k )*factor
                ncount += 1

    return p


def NormalisedJacobiTet(int C,x):
    """Computes the orthogonal base of 3D polynomials of degree less than
        or equal to n at the point x=(xi,eta,zeta) in the reference tetrahedra"""

    cdef:
        double xi = x[0] 
        double eta = x[1] 
        double zeta = x[2]
        double r,s,t

    if (eta+zeta)==0.:
        r = -1.
        s=1.
    elif zeta==1.:
        r = -1. 
        s = 1.  # or s=-1 (check that nothing changes)
    else:
        r = -2.*(1+xi)/(eta+zeta)-1.;
        s = 2.*(1+eta)/(1-zeta)-1.;
    t = zeta

    return NormalisedJacobi3D(C,[r,s,t])
    # return NormalisedJacobi3D_Native(C,[r,s,t])


@boundscheck(False)
def GradNormalisedJacobiTet(int C,x,EvalOpt=0):
    """Computes the orthogonal base of 3D polynomials of degree less than
        or equal to n at the point x=(r,s,t) in [-1,1]**3"""

    cdef:
        int N = (C+2)*(C+3)*(C+4)/6
        double[::1] p = np.zeros(N)
        double[::1] dp_dxi   = np.zeros(N)
        double[::1] dp_deta  = np.zeros(N)
        double [::1] dp_dzeta = np.zeros(N)
        double r = x[0] 
        double s = x[1]
        double t = x[2]
        double xi, eta, zeta, dr_dxi, dr_deta, dr_dzeta, ds_deta, ds_dzeta
        double p_i, p_j, p_k, dp_i, dp_j, dq_i, dq_j
        double dp_dr, dp_ds, dp_dt
        int nDeg, i, j, k
        double factor
        double eta_zeta

    # THIS MAY RUIN THE CONVERGENCE, BUT FOR POST PROCESSING ITS FINE
    if EvalOpt==1:
        if t==1.:
            t=0.99999999999
        if np.isclose(s,1.):
            s=0.99999999999

    if np.isclose(s,1.):
            s=0.999999999999

    eta = (1./2.)*(s-s*t-1.-t)
    xi = -(1./2.)*(r+1)*(eta+t)-1.
    zeta = 1.0*t
    
    # THIS MAY RUIN THE CONVERGENCE, BUT FOR POST PROCESSING ITS FINE
    if eta == 0. and zeta == 0.:
        eta = 1.0e-14
        zeta = 1e-14

    eta_zeta = eta+zeta
    if np.isclose(eta_zeta,0.):
        eta_zeta = 0.000000001
    # dr_dxi   = -2./(eta+zeta)
    # dr_deta  = 2.*(1.+xi)/(eta+zeta)**2
    dr_dxi   = -2./eta_zeta
    dr_deta  = 2.*(1.+xi)/eta_zeta**2
    dr_dzeta = dr_deta

    ds_deta  = 2./(1.-zeta)
    ds_dzeta = 2.*(1.+eta)/(1.-zeta)**2

    # DERIVATIVE OF t IS NOT REQUIRED AS t=zeta


    # ORDERING: FIRST INCREASING THE DEGREE AND THEN LEXOGRAPHIC ORDER
    cdef int ncount = 0
    # LOOP ON DEGREE
    for nDeg in range(0,C+2):
        # LOOP INCREASING i
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1.
                q_i = 1.
                dp_i = 0.
                dq_i = 0.
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1]
                dp_i = JacobiPolynomials(i-1,r,1.,1.)[-1]*(i+1.)/2.

                q_i = q_i*(1.-s)/2.
                dq_i = q_i*(-i)/(1.-s)
            # LOOP INCREASING j
            for j in range(0,nDeg-i+1):
                if j==0:
                    p_j = 1.
                    q_j = ((1.-t)/2.)**i
                    dp_j = 0.
                    dq_j = q_j*(-(i+j))/(1.-t)  
                else:
                    p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1]
                    dp_j = JacobiPolynomials(j-1,s,2.*i+2.,1.)[-1]*(j+2.*i+2.)/2.  
                    
                    q_j = q_j*(1.-t)/2.
                    dq_j = q_j*(-(i+j))/(1.-t)

                # Value for k
                k = nDeg-(i+j);
                if k==0:
                    p_k = 1.;  dp_k = 0.; 
                else:
                    p_k = JacobiPolynomials(k,t,2.*(i+j)+2.,0.)[-1]
                    dp_k = JacobiPolynomials(k-1,t,2.*(i+j)+3.,1.)[-1]*(k+2.*i+2.*j+3.)/2.
        
                factor = math.sqrt( (2.*i+1.)*(i+j+1.)*(2.*(i+j+k)+3.)/4. )
                # NORMALISED POLYNOMIAL
                p[ncount] = ( p_i*q_i*p_j*q_j*p_k )*factor
                # Derivatives with respect to (r,s,t)
                dp_dr = ( (dp_i)*q_i*p_j*q_j*p_k )*factor 
                dp_ds = ( p_i*(dq_i*p_j+q_i*dp_j)*q_j*p_k )*factor
                dp_dt = ( p_i*q_i*p_j*(dq_j*p_k+q_j*dp_k) )*factor
                # Derivatives with respect to (xi,eta,zeta)
                dp_dxi[ncount]   = dp_dr*dr_dxi
                dp_deta[ncount]  = dp_dr*dr_deta + dp_ds*ds_deta
                dp_dzeta[ncount] = dp_dr*dr_dzeta + dp_ds*ds_dzeta + dp_dt

                ncount += 1

    return p,dp_dxi,dp_deta,dp_dzeta




from __future__ import print_function, division
import numpy as np
from scipy.sparse import coo_matrix, csc_matrix, lil_matrix


def AssemblyBEM2D(C, global_coord, boundary_elements, element_connectivity, dN, Basis, w, z, Jacobian, nx, ny, XCO, YCO, geo_args):

    # Allocate the two kernels
    stiffness_K1 = np.zeros((global_coord.shape[0],global_coord.shape[0]))
    stiffness_K2 = np.zeros((global_coord.shape[0],global_coord.shape[0]))

    # Loop over collocation points
    for j in range(0,global_coord.shape[0]):
        XP = global_coord[j,0];     YP = global_coord[j,1]
        # Loop over elements
        for elem in range(0,boundary_elements.shape[0]):
            # Carry out usual Gaussian integration
            for g in range(0,w.shape[0]):
                # Compute the radial distance
                RA = np.sqrt((XCO[g,elem]-XP)**2+(YCO[g,elem]-YP)**2)
                # Compute Kernels - Assuming both sides are multiplied by 2pi
                K1 = (-1.0/(RA**2))*((XP-XCO[g,elem])*nx[g,elem]+(YP-YCO[g,elem])*ny[g,elem])
                # K2 = np.log(1.0/RA)
                K2 = -np.log(RA)
                # Fill Kernel Matrices
                stiffness_K1[j,element_connectivity[elem,0:]]+= K1*Basis[0:,g]*Jacobian[g,elem]*w[g]
                stiffness_K2[j,element_connectivity[elem,0:]]+= K2*Basis[0:,g]*Jacobian[g,elem]*w[g]


    # Implicit integration for diagonal components of K1
    for cols in range(0,stiffness_K1.shape[0]):
        stiffness_K1[cols,cols] = 1.0*(0-np.sum(stiffness_K1[cols,0:])-stiffness_K1[cols,cols])

    # Make modified stiffness matrix
    mod_stiffness_K1 = np.zeros((stiffness_K1.shape[0]+4,stiffness_K1.shape[1]+4))
    mod_stiffness_K2 = np.zeros((stiffness_K2.shape[0]+4,stiffness_K2.shape[1]+4))

    if geo_args.Lagrange_Multipliers == 'activated':
        # Normal BE blocks
        mod_stiffness_K1[0:global_coord.shape[0],0:global_coord.shape[0]] = stiffness_K1
        mod_stiffness_K2[0:global_coord.shape[0],0:global_coord.shape[0]] = stiffness_K2

        # Off diagonal blocks
        for i in range(0,global_coord.shape[0]):
            if global_coord[i,2]==1 or global_coord[i,2]==2 or global_coord[i,2]==3 or global_coord[i,2]==4:
                mod_stiffness_K1[i,global_coord.shape[0]+np.abs(global_coord[i,2])-1] = 1
            if global_coord[i,2]==-1 or global_coord[i,2]==-2 or global_coord[i,2]==-3 or global_coord[i,2]==-4:
                mod_stiffness_K1[i,global_coord.shape[0]+np.abs(global_coord[i,2])-1] = -1

        mod_stiffness_K1[global_coord.shape[0]:,0:]=mod_stiffness_K1[0:,global_coord.shape[0]:].T

        stiffness_K1 = mod_stiffness_K1
        stiffness_K2 = mod_stiffness_K2


    # # Make dense matrix a sparse matrix as sparse assembly is not efficient
    # stiffness_K1_sparse = lil_matrix((stiffness_K1.shape[0],stiffness_K1.shape[1]))
    # stiffness_K2_sparse = lil_matrix((stiffness_K1.shape[0],stiffness_K1.shape[1]))
    # # kk[:,0]=stiffness_K1[:,0]
    # for i in range(0,16):
    #   for j in range(0,16):
    #       stiffness_K1_sparse[i,j] = stiffness_K1[i,j]
    #       stiffness_K2_sparse[i,j] = stiffness_K2[i,j]


    return stiffness_K1, stiffness_K2






def AssemblyBEM2D_Sparse(C,global_coord,boundary_elements,element_connectivity,dN,Basis,w,z,Jacobian,nx,ny,XCO,YCO,geo_args):


    I_k1 = np.zeros((global_coord.shape[0]*global_coord.shape[0])); J_k1 = np.zeros((global_coord.shape[0]*global_coord.shape[0]));
    V_k1 = np.zeros((global_coord.shape[0]*global_coord.shape[0]))
    # I_k2 = np.zeros((global_coord.shape[0]*global_coord.shape[0]));   J_k2 = np.zeros((global_coord.shape[0]*global_coord.shape[0]));
    V_k2 = np.zeros((global_coord.shape[0]*global_coord.shape[0]))


    # Loop over collocation points
    for j in range(0,global_coord.shape[0]):
        XP = global_coord[j,0];     YP = global_coord[j,1]
        # Loop over elements
        for elem in range(0,boundary_elements.shape[0]):
            for k in range(0,element_connectivity.shape[1]):
                # Carry out usual Gaussian integration
                for g in range(0,w.shape[0]):
                    # Compute the radial distance
                    RA = np.sqrt((XCO[g,elem]-XP)**2+(YCO[g,elem]-YP)**2)
                    # Compute Kernels - Assuming both sides are multiplied by 2pi
                    K1 = (-1.0/(RA**2))*((XP-XCO[g,elem])*nx[g,elem]+(YP-YCO[g,elem])*ny[g,elem])
                    K2 = np.log(1.0/RA)


                    # Fill Kernel Matrices
                    I_k1[j*global_coord.shape[0]+j] = j
                    J_k1[element_connectivity[elem,k]*global_coord.shape[0]+element_connectivity[elem,k]] = element_connectivity[elem,k]
                    V_k1[j*global_coord.shape[0]+j] += K1*Basis[k,g]*Jacobian[g,elem]*w[g]
                    V_k2[j*global_coord.shape[0]+j] += K2*Basis[k,g]*Jacobian[g,elem]*w[g]


    stiffness_K1 = coo_matrix((V_k1,(I_k1,J_k1)),shape=((global_coord.shape[0],global_coord.shape[0]))).tocsc()
    stiffness_K2 = coo_matrix((V_k2,(I_k1,J_k1)),shape=((global_coord.shape[0],global_coord.shape[0]))).tocsc()


    # # Make modified stiffness matrix
    # mod_stiffness_K1 = csc_matrix((stiffness_K1.shape[0]+4,stiffness_K1.shape[1]+4))
    # mod_stiffness_K2 = csc_matrix((stiffness_K1.shape[0]+4,stiffness_K1.shape[1]+4))
    mod_stiffness_K1 = lil_matrix((stiffness_K1.shape[0]+4,stiffness_K1.shape[1]+4))
    mod_stiffness_K2 = lil_matrix((stiffness_K1.shape[0]+4,stiffness_K1.shape[1]+4))

    if geo_args.Lagrange_Multipliers=='activated':
        # Normal BE blocks
        mod_stiffness_K1[0:global_coord.shape[0],0:global_coord.shape[0]] = stiffness_K1
        mod_stiffness_K2[0:global_coord.shape[0],0:global_coord.shape[0]] = stiffness_K2

        # Off diagonal blocks
        for i in range(0,global_coord.shape[0]):
            if global_coord[i,2]==1 or global_coord[i,2]==2 or global_coord[i,2]==3 or global_coord[i,2]==4:
                mod_stiffness_K1[i,global_coord.shape[0]+np.abs(global_coord[i,2])-1] = 1
            if global_coord[i,2]==-1 or global_coord[i,2]==-2 or global_coord[i,2]==-3 or global_coord[i,2]==-4:
                mod_stiffness_K1[i,global_coord.shape[0]+np.abs(global_coord[i,2])-1] = -1

        mod_stiffness_K1[global_coord.shape[0]:,0:]=mod_stiffness_K1[0:,global_coord.shape[0]:].T

        stiffness_K1 = mod_stiffness_K1
        stiffness_K2 = mod_stiffness_K2


    return stiffness_K1, stiffness_K2





from __future__ import print_function
import numpy as np 
from Florence.QuadratureRules.FeketePointsTet import *
from Florence.QuadratureRules.EquallySpacedPoints import EquallySpacedPointsTet
from Florence.FunctionSpace.JacobiPolynomials import *
from Florence.FunctionSpace.DegenerateMappings import MapXiEtaZeta2RST

def hpBases(C, xi, eta, zeta, Transform=0, EvalOpt=0, equally_spaced=False):
    """
        Transform:                  transform to from degenrate quad
        EvalOpt:                    evaluate 1 as an approximation 0.9999999
    """

    eps = FeketePointsTet(C)
    if equally_spaced:
        eps = EquallySpacedPointsTet(C)

    N = eps.shape[0]
    # Make the Vandermonde matrix
    V = np.zeros((N,N))

    for i in range(0,N):
        x = eps[i,:]
        p = NormalisedJacobiTet(C,x)
        V[i,:] = p

    

    nsize = int((C+2.)*(C+3.)*(C+4.)/6.)

    Bases = np.zeros((nsize,1))
    gBases = np.zeros((nsize,3))

    # GradNormalisedJacobiTet IS ALWAYS CALLED WITH (HEX) R,S,T COORDINATE
    if Transform:
        # IF A TRANSFORMATION TO HEX COORDINATE IS NEEDED
        r, s, t = MapXiEtaZeta2RST(xi,eta,zeta)
        p, dp_dxi, dp_deta, dp_dzeta = GradNormalisedJacobiTet(C,np.array([r,s,t]),EvalOpt)
    else:
        # IF XI,ETA,ZETA ARE DIRECTLY GIVEN IN HEX FORMAT
        p, dp_dxi, dp_deta, dp_dzeta = GradNormalisedJacobiTet(C,np.array([xi,eta,zeta]),EvalOpt)

    # ACTUAL WAY
    # Bases = np.linalg.solve(V.T,p)
    # gBases[:,0] = np.linalg.solve(V.T,dp_dxi)
    # gBases[:,1] = np.linalg.solve(V.T,dp_deta)
    # gBases[:,2] = np.linalg.solve(V.T,dp_dzeta)

    # SOLVE FOR MULTIPLE RIGHT HAND SIDES
    # ps = np.concatenate((p[:,None],dp_dxi[:,None],dp_deta[:,None],dp_dzeta[:,None]),axis=1)
    ps = np.concatenate((p,dp_dxi,dp_deta,dp_dzeta)).reshape(4,p.shape[0]).T
    tup = np.linalg.solve(V.T,ps)
    Bases = tup[:,0]
    gBases[:,0] = tup[:,1]
    gBases[:,1] = tup[:,2]
    gBases[:,2] = tup[:,3]


    return Bases, gBases



def GradhpBases(C,r,s):
    print('For nodal tetrahedral bases gradient of bases is computed within the hpBases itself')


import numpy as np
from Florence.FunctionSpace.OneDimensional import Line as OneD
from Florence.MeshGeneration.NodeArrangement import NodeArrangementQuad


def LagrangeGaussLobatto(C,zeta,eta,arrange=1):
    """Computes stable higher order Lagrangian bases with Gauss-Lobatto-Legendre points
        Refer to: Spencer's Spectral hp elements for details
    """

    Bases = np.zeros(((C+2)**2,1))
    Neta = np.zeros((C+2,1));   Nzeta = np.zeros((C+2,1))
    # Compute each from one-dimensional bases
    Nzeta[:,0] = OneD.LagrangeGaussLobatto(C,zeta)[0]
    Neta[:,0] =  OneD.LagrangeGaussLobatto(C,eta)[0]
    # Ternsorial product
    if arrange==1:
        # # Arrange in counterclockwise - THIS FUNCTION NEEDS TO BE OPTIMISED
        # zeta_index, eta_index = GetCounterClockwiseIndices(C)
        # TBases = np.dot(Nzeta,Neta.T)
        # for i in range(0,(C+2)**2):
            # Bases[i] = TBases[zeta_index[i],eta_index[i]]

        node_arranger = NodeArrangementQuad(C)[2]
        # Bases = np.dot(Nzeta,Neta.T).flatten()
        Bases = np.dot(Neta,Nzeta.T).flatten()
        Bases = Bases[node_arranger]
        Bases = Bases[:,None]

    elif arrange==0:
        # Bases[:,0] = np.dot(Nzeta,Neta.T).flatten()
        Bases[:,0] = np.dot(Neta,Nzeta.T).flatten()

    # check = np.array([
        # 0.25*(1-zeta)*(1-eta),
        # 0.25*(1+zeta)*(1-eta),
        # 0.25*(1+zeta)*(1+eta),
        # 0.25*(1-zeta)*(1+eta)
        # ])
    # print check


    return Bases

def GradLagrangeGaussLobatto(C,zeta,eta,arrange=1):
    """Computes gradients of stable higher order Lagrangian bases with Gauss-Lobatto-Legendre points
        Refer to: Spencer's Spectral hp elements for details
    """

    gBases = np.zeros(((C+2)**2,2))
    Nzeta = np.zeros((C+2,1));  Neta = np.zeros((C+2,1))
    gNzeta = np.zeros((C+2,1)); gNeta = np.zeros((C+2,1))
    # Compute each from one-dimensional bases
    Nzeta[:,0] = OneD.LagrangeGaussLobatto(C,zeta)[0]
    Neta[:,0] = OneD.LagrangeGaussLobatto(C,eta)[0]
    gNzeta[:,0] = OneD.LagrangeGaussLobatto(C,zeta)[1]
    gNeta[:,0] = OneD.LagrangeGaussLobatto(C,eta)[1]
    # Ternsorial product
    if arrange==1:

        # # Arrange counterclockwise
        # zeta_index, eta_index = GetCounterClockwiseIndices(C)
        # gTBases0 = np.dot(gNzeta,Neta.T)
        # gTBases1 = np.dot(Nzeta,gNeta.T)

        # for i in range(0,(C+2)**2):
        #     gBases[i,0] = gTBases0[zeta_index[i],eta_index[i]]
        #     gBases[i,1] = gTBases1[zeta_index[i],eta_index[i]]

        node_arranger = NodeArrangementQuad(C)[2]
        # g0 = np.dot(gNzeta,Neta.T).flatten()
        # g1 = np.dot(Nzeta,gNeta.T).flatten()
        g0 = np.dot(Nzeta,gNeta.T).flatten()
        g1 = np.dot(gNzeta,Neta.T).flatten()
        # g0 = np.dot(gNeta,Nzeta.T).flatten()
        # g1 = np.dot(Neta,gNzeta.T).flatten()
        gBases[:,0] = g0[node_arranger]
        gBases[:,1] = g1[node_arranger]

    elif arrange==0:
        # gBases[:,0] = np.dot(gNzeta,Neta.T).reshape((C+2)**2)
        # gBases[:,1] = np.dot(Nzeta,gNeta.T).reshape((C+2)**2)
        gBases[:,0] = np.dot(gNeta,Nzeta.T).reshape((C+2)**2)
        gBases[:,1] = np.dot(Neta,gNzeta.T).reshape((C+2)**2)


    # check =  0.25*np.array([[eta-1.,1-eta,1+eta,-1.-eta],[zeta-1.,-zeta-1.,1+zeta,1-zeta]])


    return gBases


import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material
from Florence.LegendreTransform import LegendreTransform


class IsotropicElectroMechanics_103(Material):
    """ Electromechanical model in terms of internal energy 
                    W(C,D) = W_neo(C) + 1/2/eps_1 (D0*D0) + 1/2/eps_2 (FD0*FD0)
    """
    
    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IsotropicElectroMechanics_103, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = self.ndim+1
        self.energy_type = "internal_energy"
        self.legendre_transform = LegendreTransform()
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"  

        if self.ndim == 2:
            self.H_VoigtSize = 5
        elif self.ndim == 3:
            self.H_VoigtSize = 9

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eps_1 = self.eps_1
        eps_2 = self.eps_2

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        D  = ElectricDisplacementx.reshape(self.ndim,1)
        Dx = D.reshape(self.ndim)
        DD = np.dot(D.T,D)[0,0]

        self.elasticity_tensor = lamb*(2.*J-1.)*einsum("ij,kl",I,I) + \
        (mu/J - lamb*(J-1))*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) )

        self.coupling_tensor = J/eps_2*(einsum('ik,j',I,Dx) + einsum('i,jk',Dx,I))

        self.dielectric_tensor = J/eps_1*np.linalg.inv(b)  + J/eps_2*I 

        # TRANSFORM TENSORS TO THEIR ENTHALPY COUNTERPART
        E_Voigt, P_Voigt, C_Voigt = self.legendre_transform.InternalEnergyToEnthalpy(self.dielectric_tensor, 
            self.coupling_tensor, self.elasticity_tensor)
        
        # BUILD HESSIAN
        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*P_Voigt),axis=1)
        H2 = np.concatenate((factor*P_Voigt.T,E_Voigt),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)

        return H_Voigt


    def CauchyStress(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eps_2 = self.eps_2

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        D = ElectricDisplacementx.reshape(self.ndim,1)

        simga_mech = 1.0*mu/J*(b - I) + lamb*(J-1)*I
        sigma_electric = J/eps_2*np.dot(D,D.T)
        sigma = simga_mech + sigma_electric

        return sigma


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        D = self.legendre_transform.GetElectricDisplacement(self, StrainTensors, ElectricFieldx, elem, gcounter)


        # SANITY CHECK FOR IMPLICIT COMPUTATUTAION OF D
        # I = StrainTensors['I']
        # J = StrainTensors['J'][gcounter]
        # b = StrainTensors['b'][gcounter]
        # E = ElectricFieldx.reshape(self.ndim,1)
        # eps_1 = self.eps_1
        # eps_2 = self.eps_2
        # inverse = np.linalg.inv(J/eps_1*np.linalg.inv(b) + J/eps_2*I)
        # D_exact = np.dot(inverse,E)
        # print np.linalg.norm(D - D_exact)
        # # return D_exact

        return D



from cython cimport boundscheck, wraparound
import numpy as np
cimport numpy as np 
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy


ctypedef np.int32_t Int # for I, J storage indices
ctypedef np.float64_t Float

# The following fused typedefs are provided for
# cases when assembly (LHS) matrix is lowered to 32 bit, but their
# performance is terrible. Recompile instead for 32bit LHSes 
# ctypedef fused Int:
#     np.int32_t
#     np.int64_t

# ctypedef fused Float:
#     np.float32_t
#     np.float64_t

@boundscheck(False)
@wraparound(False)
def SparseAssemblyNative(np.ndarray[long] i, np.ndarray[long] j,
    np.ndarray[double] coeff, np.ndarray[Int] I, np.ndarray[Int] J,
    np.ndarray[Float] V, Int elem, Int nvar, Int nodeperelem,
    np.ndarray[unsigned long,ndim=2, mode='c'] elements):

    cdef long i_shape = i.shape[0]
    cdef long j_shape = j.shape[0]
    assert i_shape==j_shape

    SparseAssemblyNative_(&i[0],&j[0],&coeff[0],&I[0],&J[0],&V[0],
        elem,nvar,nodeperelem,&elements[0,0],i_shape,j_shape)




cdef void SparseAssemblyNative_(const long *i, const long *j, const double *coeff, Int *I, Int *J,
    Float *V, Int elem, Int nvar, Int nodeperelem, const unsigned long *elements,long i_shape, long j_shape) nogil:

    cdef long *current_row_column = <long*>malloc(sizeof(long)*nvar*nodeperelem)
    cdef long *full_current_row = <long*>malloc(sizeof(long)*i_shape)
    cdef long *full_current_column = <long*>malloc(sizeof(long)*j_shape)

    cdef long iterator, counter, ncounter
    # sqrt of local capacity to be precise 
    cdef Int local_capacity = nvar*nodeperelem 

    cdef long const_elem_retriever 
    for counter in range(nodeperelem):
        const_elem_retriever = nvar*elements[elem*nodeperelem+counter]
        for ncounter in range(nvar):
            current_row_column[nvar*counter+ncounter] = const_elem_retriever+ncounter

    memcpy(full_current_row,i,i_shape*sizeof(long))
    memcpy(full_current_column,j,j_shape*sizeof(long))

    cdef long const_I_retriever 
    for counter in range(local_capacity):
        const_I_retriever = current_row_column[counter]
        for iterator in range(local_capacity):
            full_current_row[counter*local_capacity+iterator]    = const_I_retriever
            full_current_column[counter*local_capacity+iterator] = current_row_column[iterator]


    # STORE INDICES AND COEFFICIENTS IN I, J AND V VECTORS
    cdef long low, high
    low = local_capacity*local_capacity*elem
    high = local_capacity*local_capacity*(elem+1)

    cdef long incrementer = 0
    for counter in range(low,high):
        I[counter] = full_current_row[incrementer]
        J[counter] = full_current_column[incrementer]
        V[counter] = coeff[incrementer]

        incrementer += 1


    free(full_current_row)
    free(full_current_column)
    free(current_row_column)



import numpy as np 

# MAPPING FROM QUAD TO TRI
def MapRS2XiEta(r,s):
	# Given a map (r,s) in [-1,1]^2 quad, computes the mapped (xi,eta)  

	xi = 0.5*(1.+r)*(1.-s) - 1.
	eta = 1.0*s
	return xi, eta
	

# MAPPING FROM HEX TO TET
def MapRST2XiEtaZeta(r,s,t):
	# Given a map (r,s,t) in [-1,1]^3 hex, computes the mapped (xi,eta,zeta)  
	
	zeta = 1.0*t 
	eta = 0.5*(1.+s)*(1.-t) - 1.
	xi = -0.5*(1.+r)*(eta+zeta) - 1

	return xi, eta, zeta 



# MAPPING FROM TRI TO QUAD
def MapXiEta2RS(xi,eta):

	if 'array' in str(type(eta)):
		xi = xi.reshape(xi.size); eta = eta.reshape(eta.size)
		xi = np.array([xi]); eta = np.array([eta])
	else:
		xi = np.array([[xi]]); eta = np.array([[eta]])
	xi=xi.reshape(xi.shape[1])
	eta=eta.reshape(eta.shape[1])

	dum = np.linspace(0,eta.shape[0]-1,eta.shape[0])

	posSingular = np.where(eta>0.999)[0]
	posNoSingular = np.delete(dum, posSingular)
	posNoSingular = posNoSingular.astype(int)

	# Allocate
	r = np.zeros(dum.shape[0]); s = np.copy(r)
	# Fill now
	r[posSingular] = -1.
	r[posNoSingular] = 2.*(1.+xi[posNoSingular] )/(1.- eta[posNoSingular]) - 1.
	s = eta

	return r, s


# MAPPING FROM TET TO HEX
def MapXiEtaZeta2RST(xi,eta,zeta):

	if 'array' in str(type(eta)):
		xi = xi.reshape(xi.size); eta = eta.reshape(eta.size); zeta = zeta.reshape(zeta.size); 
		xi = np.array([xi]); eta = np.array([eta]); zeta = np.array([zeta]); 
	else:
		xi = np.array([[xi]]); eta = np.array([[eta]]); zeta = np.array([[zeta]])
	xi=xi.reshape(xi.shape[1])
	eta=eta.reshape(eta.shape[1])
	zeta=zeta.reshape(zeta.shape[1])

	dum = np.linspace(0,eta.shape[0]-1,eta.shape[0])


	posSingular1 = np.where(np.abs(eta+zeta)<0.001)
	posSingular2 = np.where(zeta>0.999)
	posNoSingular1 = np.delete(dum, posSingular1)
	posNoSingular1 = posNoSingular1.astype(int)
	posNoSingular2 = np.delete(dum, posSingular2)
	posNoSingular2 = posNoSingular2.astype(int)

	r = np.zeros(dum.shape[0]); s = np.copy(r); t=np.copy(r)
	# Fill now
	r[posSingular1] = -1.
	s[posSingular2] = 1.
	r[posNoSingular1] = -2.*(1.+xi[posNoSingular1])/(eta[posNoSingular1]+zeta[posNoSingular1]) - 1.
	s[posNoSingular2] = 2.*(1.+eta[posNoSingular2])/(1.-zeta[posNoSingular2]) - 1.
	t = zeta

	return r, s, t 


import numpy as np

def GetInteriorNodesCoordinates(xycoord, element_type, elem, eps, Neval):

    xycoord_higher              = np.zeros((eps.shape[0],xycoord.shape[1]))

    if element_type == 'hex':
        xycoord_higher[:8,:]    = xycoord
        xycoord_higher[8:,:]    = np.dot(Neval[:,8:].T,xycoord)

    elif element_type == 'tet':
        xycoord_higher[:4,:]    = xycoord
        xycoord_higher[4:,:]    = np.dot(Neval[:,4:].T,xycoord)

    elif element_type == 'quad':
        xycoord_higher[:4,:]    = xycoord
        xycoord_higher[4:,:]    = np.dot(Neval[:,4:].T,xycoord)

    elif element_type == 'tri':
        xycoord_higher[:3,:]    = xycoord
        xycoord_higher[3:,:]    = np.dot(Neval[:,3:].T,xycoord)

    elif element_type == 'line':
        xycoord_higher[:2,:]    = xycoord
        xycoord_higher[2:,:]    = np.dot(Neval[:,2:].T,xycoord)


    return xycoord_higher


# This program generates Jacobi polynomials of any order used for generating interpolation functions
# Based on the book "George Em Karniadakis & Spencer J. Sherwin (1999) - Spectral hp-element methods for CFD" Appendix A
# from WrittenJacobiPolynomials import WrittenJacobiPolynomials
# from WrittenJacobiPolynomials_8 import WrittenJacobiPolynomials_8
# import JacobiPolynomials_Cy

# from numba import jit
# from numba.decorators import jit, autojit
# import numpy as np 

# @jit
def JacobiPolynomials(n,xi,a=0,b=0):
    # Input arguments:
    # n - polynomial degree
    # xi - evalution point
    # a,b - alpha and beta parameters for Jacobi Polynmials
            # a=b=0 for Legendre polynomials
            # a=b=-0.5 for Chebychev polynomials

    # Written Jacobi is not a good idea at least for Python (Numpy/Scipy)
    # P = []
    # # if n < 17:
    # if n < 50:
    #   # P = WrittenJacobiPolynomials(n,xi,a,b)
    #   P = JacobiPolynomials_Cy.JacobiPolynomials(n,xi,a,b)
    # else:

    # The first two polynomials
    # P = np.zeros((n+1,1))
    P=[0]*(n+1)  # List seems much faster than np.array here

    P[0] = 1.0
    if n>0:
        P[1] = 0.5*((a-b)+(a+b+2)*xi)

    if n>1:
        for p in range(1,n):
            # Evaluate coefficients
            a1n = 2*(p+1)*(p+a+b+1)*(2*p+a+b)
            a2n = (2*p+a+b+1)*(a**2-b**2)
            a3n = (2*p+a+b)*(2*p+a+b+1)*(2*p+a+b+2)
            a4n = 2*(p+a)*(p+b)*(2*p+a+b+2)
            # print p
            P[p+1] = ((a2n+a3n*xi)*P[p]-a4n*P[p-1])/a1n

    return P

# @jit
def DiffJacobiPolynomials(n,xi,a=0,b=0,opt=0):
    # opt is for Gauss-Lobatto integration purpose only
    # Compute derivatives
    # dP = np.zeros((n+1,1))
    dP=[0]*(n+1)    # List seems much faster than np.array here 

    if opt==1:
        P = JacobiPolynomials(n,xi,a+1,b+1)
    else:
        P = JacobiPolynomials(n,xi,a,b)

    for p in range(1,n+1):
        dP[p] = 0.5*(a+b+p+1)*P[p-1]
        
    return dP




import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt

class RegularisedNeoHookean(Material):
    """The fundamental Neo-Hookean internal energy, described in Bonet et. al.

        W(C) = mu/2*(C:I-3)- mu*lnJ + lamb/2*(J-1)**2 + W_elastic

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(RegularisedNeoHookean, self).__init__(mtype, ndim, **kwargs)
        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx=0, elem=0):
        return


    def Hessian(self,StrainTensors,ElectricFieldx=None,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        # Jb = 1.27168554933
        Jb = self.Jbar[elem]

        mu2 = self.mu/J- self.lamb*(J-Jb)
        lamb2 = self.lamb*(2*J-Jb)

        H_Voigt_n = lamb2*self.vIijIkl+mu2*self.vIikIjl

        # return H_Voigt

        H_Voigt_l = Voigt(self.lamb*einsum('ij,kl',I,I)+self.mu*(einsum('ik,jl',I,I)+einsum('il,jk',I,I)) ,1)

        alp = 0.1
        # H_Voigt = (1-alp)*H_Voigt_l + alp*H_Voigt_n
        H_Voigt = H_Voigt_n
        # return H_Voigt_l + H_Voigt_n
        return H_Voigt

    def CauchyStress(self,StrainTensors,ElectricFieldx=None,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        strain = StrainTensors['strain'][gcounter]

        # mu = self.mu
        # lamb = self.lamb

        # return 1.0*mu/J*b + (lamb*(J-1.0)-mu/J)*I

        mu = self.mu
        lamb = self.lamb
        # Jb = 1.27168554933
        Jb = self.Jbar[elem]

        # CHECK IF THIS IS NECESSARY
        if self.ndim == 3:
            tre = trace(strain)
        elif self.ndim == 2:
            tre = trace(strain) + 1

        sigma_linear =  2.*mu*strain + lamb*tre*I
        simga_nonlinear = 1.0*mu/J*b + (lamb*(J-Jb)-mu/J)*I

        alp = 0.1
        # simga =  (1-alp)*simga_nonlinear + alp*sigma_linear
        simga = simga_nonlinear

        return simga



